"""Get the LLM client."""
from enum import Enum
from cognee.config import Config
from .anthropic.adapter import AnthropicAdapter
from .openai.adapter import OpenAIAdapter
from .generic_llm_api.adapter import GenericAPIAdapter

# Define an Enum for LLM Providers
class LLMProvider(Enum):
    OPENAI = "openai"
    OLLAMA = "ollama"
    ANTHROPIC = "anthropic"
    CUSTOM = "custom"

config = Config()
config.load()

def get_llm_client():
    """Get the LLM client based on the configuration using Enums."""
    provider = LLMProvider(config.llm_provider)

    if provider == LLMProvider.OPENAI:
        return OpenAIAdapter(config.openai_key, config.openai_model)
    elif provider == LLMProvider.OLLAMA:
        return GenericAPIAdapter(config.ollama_endpoint, config.ollama_key, config.ollama_model)
    elif provider == LLMProvider.ANTHROPIC:
        return AnthropicAdapter(config.custom_model)
    elif provider == LLMProvider.CUSTOM:
        return GenericAPIAdapter(config.custom_endpoint, config.custom_key, config.custom_model)
    else:
        raise ValueError(f"Unsupported LLM provider: {provider}")



""" This module contains utility functions for the cognee. """

import os
import graphistry
import networkx as nx
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import tiktoken
import nltk
from posthog import Posthog

from cognee.config import Config
import uuid
import datetime

config = Config()
config.load()


def send_telemetry( posthog, event_name="COGNEE_ADD"):
    if os.getenv("TELEMETRY_DISABLED"):
        return

    posthog = Posthog(project_api_key='phc_bbR86N876kwub62Lr3dhQ7zIeRyMMMm0fxXqxPqzLm3',
                      host='https://eu.i.posthog.com')

    user_id = str(uuid.uuid4())
    current_time = datetime.datetime.now()
    properties = {
        "time": current_time.strftime('%m/%d/%Y')

    }

    try:
        posthog.capture(user_id, event_name, properties)
    except Exception as e:
        print('ERROR sending telemetric data to Posthog. See exception: %s', e)
def get_document_names(doc_input):
    """
    Get a list of document names.

    This function takes doc_input, which can be a folder path,
    a single document file path, or a document name as a string.
    It returns a list of document names based on the doc_input.

    Args:
        doc_input (str): The doc_input can be a folder path, a single document file path,
        or a document name as a string.

    Returns:
        list: A list of document names.

    Example usage:
        - Folder path: get_document_names(".data")
        - Single document file path: get_document_names(".data/example.pdf")
        - Document name provided as a string: get_document_names("example.docx")

    """
    if isinstance(doc_input, list):
        return doc_input
    if os.path.isdir(doc_input):
        # doc_input is a folder
        folder_path = doc_input
        document_names = []
        for filename in os.listdir(folder_path):
            if os.path.isfile(os.path.join(folder_path, filename)):
                document_names.append(filename)
        return document_names
    elif os.path.isfile(doc_input):
        # doc_input is a single document file
        return [os.path.basename(doc_input)]
    elif isinstance(doc_input, str):
        # doc_input is a document name provided as a string
        return [doc_input]
    else:
        # doc_input is not valid
        return []

def num_tokens_from_string(string: str, encoding_name: str) -> int:
    """Returns the number of tokens in a text string."""

    # tiktoken.get_encoding("cl100k_base")
    encoding = tiktoken.encoding_for_model(encoding_name)
    num_tokens = len(encoding.encode(string))
    return num_tokens


def trim_text_to_max_tokens(text: str, max_tokens: int, encoding_name: str) -> str:
    """
    Trims the text so that the number of tokens does not exceed max_tokens.

    Args:
    text (str): Original text string to be trimmed.
    max_tokens (int): Maximum number of tokens allowed.
    encoding_name (str): The name of the token encoding to use.

    Returns:
    str: Trimmed version of text or original text if under the limit.
    """
    # First check the number of tokens
    num_tokens = num_tokens_from_string(text, encoding_name)

    # If the number of tokens is within the limit, return the text as is
    if num_tokens <= max_tokens:
        return text

    # If the number exceeds the limit, trim the text
    # This is a simple trim, it may cut words in half; consider using word boundaries for a cleaner cut
    encoded_text = tiktoken.get_encoding(encoding_name).encode(text)
    trimmed_encoded_text = encoded_text[:max_tokens]
    # Decoding the trimmed text
    trimmed_text = tiktoken.get_encoding(encoding_name).decode(trimmed_encoded_text)
    return trimmed_text


def format_dict(d):
    """Format a dictionary as a string."""
    # Initialize an empty list to store formatted items
    formatted_items = []

    # Iterate through all key-value pairs
    for key, value in d.items():
        # Format key-value pairs with a colon and space, and adding quotes for string values
        formatted_item = (
            f"{key}: '{value}'" if isinstance(value, str) else f"{key}: {value}"
        )
        formatted_items.append(formatted_item)

    # Join all formatted items with a comma and a space
    formatted_string = ", ".join(formatted_items)

    # Add curly braces to mimic a dictionary
    formatted_string = f"{{{formatted_string}}}"

    return formatted_string


def generate_color_palette(unique_layers):
    colormap = plt.cm.get_cmap("viridis", len(unique_layers))
    colors = [colormap(i) for i in range(len(unique_layers))]
    hex_colors = ["#%02x%02x%02x" % (int(rgb[0]*255), int(rgb[1]*255), int(rgb[2]*255)) for rgb in colors]

    return dict(zip(unique_layers, hex_colors))


async def register_graphistry():
    graphistry.register(api = 3, username = config.graphistry_username, password = config.graphistry_password)


def prepare_edges(graph):
    return nx.to_pandas_edgelist(graph)


def prepare_nodes(graph, include_size=False):
    nodes_data = []
    for node in graph.nodes:
        node_info = graph.nodes[node]
        description = node_info.get("layer_description", {}).get("layer", "Default Layer") if isinstance(
            node_info.get("layer_description"), dict) else node_info.get("layer_description", "Default Layer")
        # description = node_info['layer_description']['layer'] if isinstance(node_info.get('layer_description'), dict) and 'layer' in node_info['layer_description'] else node_info.get('layer_description', node)
        # if isinstance(node_info.get('layer_description'), dict) and 'layer' in node_info.get('layer_description'):
        #     description = node_info['layer_description']['layer']
        # # Use 'layer_description' directly if it's not a dictionary, otherwise default to node ID
        # else:
        #     description = node_info.get('layer_description', node)

        node_data = {"id": node, "layer_description": description}
        if include_size:
            default_size = 10  # Default node size
            larger_size = 20  # Size for nodes with specific keywords in their ID
            keywords = ["DOCUMENT", "User", "LAYER"]
            node_size = larger_size if any(keyword in str(node) for keyword in keywords) else default_size
            node_data["size"] = node_size
        nodes_data.append(node_data)

    return pd.DataFrame(nodes_data)


async def render_graph(graph, include_nodes=False, include_color=False, include_size=False, include_labels=False):
    await register_graphistry()
    edges = prepare_edges(graph)
    plotter = graphistry.edges(edges, "source", "target")

    if include_nodes:
        nodes = prepare_nodes(graph, include_size=include_size)
        plotter = plotter.nodes(nodes, "id")


        if include_size:
            plotter = plotter.bind(point_size="size")


        if include_color:
            unique_layers = nodes["layer_description"].unique()
            color_palette = generate_color_palette(unique_layers)
            plotter = plotter.encode_point_color("layer_description", categorical_mapping=color_palette,
                                                 default_mapping="silver")


        if include_labels:
            plotter = plotter.bind(point_label = "layer_description")



    # Visualization
    url = plotter.plot(render=False, as_files=True, memoize=False)
    print(f"Graph is visualized at: {url}")


def sanitize_df(df):
    """Replace NaNs and infinities in a DataFrame with None, making it JSON compliant."""
    return df.replace([np.inf, -np.inf, np.nan], None)


def get_entities(tagged_tokens):
    nltk.download("maxent_ne_chunker")
    from nltk.chunk import ne_chunk
    return ne_chunk(tagged_tokens)


def extract_pos_tags(sentence):
    """Extract Part-of-Speech (POS) tags for words in a sentence."""

    # Ensure that the necessary NLTK resources are downloaded
    nltk.download("words")
    nltk.download("punkt")
    nltk.download("averaged_perceptron_tagger")

    from nltk.tag import pos_tag
    from nltk.tokenize import word_tokenize

    # Tokenize the sentence into words
    tokens = word_tokenize(sentence)

    # Tag each word with its corresponding POS tag
    pos_tags = pos_tag(tokens)

    return pos_tags


def extract_named_entities(sentence):
    """Extract Named Entities from a sentence."""
    # Tokenize the sentence into words
    tagged_tokens = extract_pos_tags(sentence)

    # Perform Named Entity Recognition (NER) on the tagged tokens
    entities = get_entities(tagged_tokens)

    return entities


def extract_sentiment_vader(text):
    """
    Analyzes the sentiment of a given text using the VADER Sentiment Intensity Analyzer.

    Parameters:
    text (str): The text to analyze.

    Returns:
    dict: A dictionary containing the polarity scores for the text.
    """
    from nltk.sentiment import SentimentIntensityAnalyzer

    nltk.download("vader_lexicon")

    # Initialize the VADER Sentiment Intensity Analyzer
    sia = SentimentIntensityAnalyzer()

    # Obtain the polarity scores for the text
    polarity_scores = sia.polarity_scores(text)

    return polarity_scores


if __name__ == "__main__":
    sample_text = "I love sunny days, but I hate the rain."
    sentiment_scores = extract_sentiment_vader(sample_text)
    print("Sentiment analysis results:", sentiment_scores)



from cognee.infrastructure.databases.vector.qdrant.adapter import CollectionConfig
from cognee.infrastructure.databases.vector.get_vector_database import get_vector_database

async def create_vector_memory(memory_name: str, collection_config: CollectionConfig):
    vector_db = get_vector_database()

    return await vector_db.create_collection(memory_name, collection_config)



from os import path
import logging
from cognee.root_dir import get_absolute_path

def read_query_prompt(prompt_file_name: str):
    """Read a query prompt from a file."""
    try:
        file_path = path.join(get_absolute_path("./infrastructure/llm/prompts"), prompt_file_name)

        with open(file_path, "r", encoding = "utf-8") as file:
            return file.read()
    except FileNotFoundError:
        logging.error(f"Error: Prompt file not found. Attempted to read: %s {file_path}")
        return None
    except Exception as e:
        logging.error(f"An error occurred: %s {e}")
        return None



import uuid

from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
from cognee.shared.data_models import GraphDBType



async def group_nodes_by_layer(node_descriptions):
    """ Group nodes by layer decomposition uuid """
    grouped_data = {}

    for item in node_descriptions:
        uuid = item["layer_decomposition_uuid"]
        if uuid not in grouped_data:
            grouped_data[uuid] = []
        grouped_data[uuid].append(item)

    print("GROUPED DATA", grouped_data)

    return grouped_data


async def get_node_by_unique_id(graph, unique_id):
    """ Get a node by its unique_id"""
    # Iterate through all nodes and their attributes in the graph
    for node, attrs in graph.nodes(data=True):
        # Check if the current node's attributes contain the unique_id we're looking for
        if attrs.get('unique_id') == unique_id:
            return node  # Return the node (identifier) if found
    return None

async def connect_nodes_in_graph(graph, relationship_dict, score_threshold=0.9):
    """ Connect nodes in the graph based on the relationship_dict and score_threshold"""
    if not graph or not relationship_dict:
        return graph

    for _, relationships in relationship_dict.items():
        for relationship in relationships:

            if relationship['score'] > score_threshold:

                # For NetworkX
                if infrastructure_config.get_config()["graph_engine"] == GraphDBType.NETWORKX:
                    searched_node_id_found = await get_node_by_unique_id(graph.graph, relationship['searched_node_id'])
                    original_id_for_search_found = await get_node_by_unique_id(graph.graph, relationship['original_id_for_search'])
                    if searched_node_id_found and original_id_for_search_found:
                        await graph.add_edge(
                            searched_node_id_found,
                            original_id_for_search_found,
                            weight=relationship['score'],
                            score_metadata=relationship.get('score_metadata', {}),
                            id = f""" SEMANTIC_CONNECTION_{searched_node_id_found}_{original_id_for_search_found}_{str(uuid.uuid4())}"""
                        )

                # For Neo4j
                elif infrastructure_config.get_config()["graph_engine"] == GraphDBType.NEO4J:
                    # Neo4j specific logic to add an edge
                    # This is just a placeholder, replace it with actual Neo4j logic
                    print("query is ", f"""MATCH (a), (b) WHERE a.unique_id = '{relationship['searched_node_id']}' AND b.unique_id = '{relationship['original_id_for_search']}' CREATE (a)-[:CONNECTED {{weight:{relationship['score']}}}]->(b)""")
                    result = await graph.query(f"""MATCH (a), (b) WHERE a.unique_id = '{relationship['searched_node_id']}' AND b.unique_id = '{relationship['original_id_for_search']}'
                              CREATE (a)-[:SEMANTIC_CONNECTION {{weight:{relationship['score']}}}]->(b)""")
                    await graph.close()


def graph_ready_output(results):
    """ Generate a dictionary of relationships from the results of the graph search."""
    relationship_dict = {}

    for result in results:
        layer_id = result["layer_id"]
        layer_nodes = result["layer_nodes"]

        # Ensure there's a list to collect related items for this uuid
        if layer_id not in relationship_dict:
            relationship_dict[layer_id] = []

        for node in layer_nodes:  # Iterate over the list of ScoredPoint lists
            for score_point in node["score_points"]:
                # Append a new dictionary to the list associated with the uuid
                relationship_dict[layer_id].append({
                    "collection_id": layer_id,
                    "searched_node_id": node["id"],
                    "score": score_point.score,
                    "score_metadata": score_point.payload,
                    "original_id_for_search": score_point.id,
                })

    return relationship_dict


if __name__ == "__main__":

    async def main():
        graph_client = await get_graph_client(GraphDBType.NEO4J)
        graph = graph_client.graph

        # for nodes, attr in graph.nodes(data=True):
        #     if 'd0bd0f6a-09e5-4308-89f6-400d66895126' in nodes:
        #         print(nodes)

        #
        # relationships = {'SuaGeKyKWKWyaSeiqWeWaSyuSKqieSamiyah': [{'collection_id': 'SuaGeKyKWKWyaSeiqWeWaSyuSKqieSamiyah', 'searched_node_id': 'd0bd0f6a-09e5-4308-89f6-400d66895126', 'score': 1.0, 'score_metadata': {'text': 'Pravilnik o izmenama i dopunama Pravilnika o sadržini, načinu i postupku izrade i način vršenja kontrole tehničke dokumentacije prema klasi i nameni objekata'}, 'original_id_for_search': '2801f7b5-55bf-499b-9843-97d48f8e067a'}, {'collection_id': 'SuaGeKyKWKWyaSeiqWeWaSyuSKqieSamiyah', 'searched_node_id': 'd0bd0f6a-09e5-4308-89f6-400d66895126', 'score': 0.1648828387260437, 'score_metadata': {'text': 'Zakon o planiranju i izgradnji'}, 'original_id_for_search': '57966b55-33e2-4eae-a7fa-2f0237643bbe'}, {'collection_id': 'SuaGeKyKWKWyaSeiqWeWaSyuSKqieSamiyah', 'searched_node_id': 'd0bd0f6a-09e5-4308-89f6-400d66895126', 'score': 0.12986786663532257, 'score_metadata': {'text': 'Službeni glasnik RS, broj 77/2015'}, 'original_id_for_search': '0f626d48-4441-43c1-9060-ea7e54f6d8e2'}, {'collection_id': 'SuaGeKyKWKWyaSeiqWeWaSyuSKqieSamiyah', 'searched_node_id': 'c9b9a460-c64a-4e2e-a4d6-aa5b3769274b', 'score': 1.0, 'score_metadata': {'text': 'Službeni glasnik RS, broj 77/2015'}, 'original_id_for_search': '0f626d48-4441-43c1-9060-ea7e54f6d8e2'}, {'collection_id': 'SuaGeKyKWKWyaSeiqWeWaSyuSKqieSamiyah', 'searched_node_id': 'c9b9a460-c64a-4e2e-a4d6-aa5b3769274b', 'score': 0.07603412866592407, 'score_metadata': {'text': 'Prof. dr Zorana Mihajlović'}, 'original_id_for_search': '5d064a62-3cd6-4895-9f60-1a0d8bc299e8'}, {'collection_id': 'SuaGeKyKWKWyaSeiqWeWaSyuSKqieSamiyah', 'searched_node_id': 'c9b9a460-c64a-4e2e-a4d6-aa5b3769274b', 'score': 0.07226034998893738, 'score_metadata': {'text': 'Ministar građevinarstva, saobraćaja i infrastrukture'}, 'original_id_for_search': 'f5d052ca-c4a0-490e-a3ac-d8ad522dea83'}, {'collection_id': 'SuaGeKyKWKWyaSeiqWeWaSyuSKqieSamiyah', 'searched_node_id': 'bbd6d2d6-e673-4b59-a50c-516972a9d0de', 'score': 0.5, 'score_metadata': {'text': 'Pravilnik o izmenama i dopunama Pravilnika o sadržini, načinu i postupku izrade i način vršenja kontrole tehničke dokumentacije prema klasi i nameni objekata'}, 'original_id_for_search': '2801f7b5-55bf-499b-9843-97d48f8e067a'}]}
        #
        # connect_nodes_in_graph(graph, relationships)

        from cognee.utils import render_graph

        graph_url = await render_graph(graph)

        print(graph_url)

    import asyncio
    asyncio.run(main())



from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.llm.prompts import read_query_prompt
from cognee.infrastructure.llm.get_llm_client import get_llm_client

async def extract_summary(content: str, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    system_prompt = read_query_prompt("summarize_content.txt")

    llm_output = await llm_client.acreate_structured_output(content, system_prompt, response_model)

    return llm_output.model_dump()



from typing import Type, List
from pydantic import BaseModel
from cognee.infrastructure.llm.prompts import read_query_prompt
from cognee.infrastructure.llm.get_llm_client import get_llm_client


async def extract_categories(content: str, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    system_prompt = read_query_prompt("classify_content.txt")

    llm_output = await llm_client.acreate_structured_output(content, system_prompt, response_model)

    return process_categories(llm_output.model_dump())

def process_categories(llm_output) -> List[dict]:
    # Extract the first subclass from the list (assuming there could be more)
    data_category = llm_output["label"]["subclass"][0]

    data_type = llm_output["label"]["type"].lower()

    return [{
        "data_type": data_type,
        # The data_category is the value of the Enum member (e.g., "News stories and blog posts")
        "category_name": data_category.value
    }]



import asyncio
import logging
from cognee.root_dir import get_absolute_path
from .extract_knowledge_graph_module import ExtractKnowledgeGraph
from .extract_content_graph import extract_content_graph

logger = logging.getLogger("extract_knowledge_graph(text: str)")

async def extract_knowledge_graph(text: str, cognitive_layer, graph_model):
    try:
        compiled_extract_knowledge_graph = ExtractKnowledgeGraph()
        compiled_extract_knowledge_graph.load(get_absolute_path("./programs/extract_knowledge_graph/extract_knowledge_graph.json"))

        event_loop = asyncio.get_event_loop()

        def sync_extract_knowledge_graph():
            return compiled_extract_knowledge_graph(context = text, question = "")

        return (await event_loop.run_in_executor(None, sync_extract_knowledge_graph)).graph
        # return compiled_extract_knowledge_graph(text, question = "").graph
    except Exception as error:
        logger.error("Error extracting graph from content: %s", error, exc_info = True)
        
        return await extract_content_graph(text, cognitive_layer, graph_model)



from cognee.config import Config
from .databases.relational import DuckDBAdapter, DatabaseEngine
from .databases.vector.vector_db_interface import VectorDBInterface
from .databases.vector.qdrant.QDrantAdapter import QDrantAdapter
from .databases.vector.embeddings.DefaultEmbeddingEngine import DefaultEmbeddingEngine
from .llm.llm_interface import LLMInterface
from .llm.openai.adapter import OpenAIAdapter
from .files.storage import LocalStorage
from .data.chunking.DefaultChunkEngine import DefaultChunkEngine
from ..shared.data_models import GraphDBType, DefaultContentPrediction, KnowledgeGraph, SummarizedContent, \
    LabeledContent, DefaultCognitiveLayer

config = Config()
config.load()

class InfrastructureConfig():
    system_root_directory: str = config.system_root_directory
    data_root_directory: str = config.data_root_directory
    llm_provider: str = config.llm_provider
    database_engine: DatabaseEngine = None
    vector_engine: VectorDBInterface = None
    graph_engine: GraphDBType = None
    llm_engine: LLMInterface = None
    classification_model = None
    summarization_model = None
    labeling_model = None
    graph_model = None
    cognitive_layer_model = None
    intra_layer_score_treshold = None
    embedding_engine = None
    connect_documents = config.connect_documents
    database_directory_path: str = None
    database_file_path: str = None
    chunk_strategy = config.chunk_strategy
    chunk_engine = None

    def get_config(self, config_entity: str = None) -> dict:
        if (config_entity is None or config_entity == "database_engine") and self.database_engine is None:
            db_path = self.system_root_directory + "/" + config.db_path

            LocalStorage.ensure_directory_exists(db_path)

            self.database_engine = DuckDBAdapter(
                db_name = config.db_name,
                db_path = db_path
            )

        if self.graph_engine is None:
            self.graph_engine = GraphDBType.NETWORKX

        if self.classification_model is None:
            self.classification_model = DefaultContentPrediction

        if self.summarization_model is None:
            self.summarization_model = SummarizedContent

        if self.labeling_model is None:
            self.labeling_model = LabeledContent

        if self.graph_model is None:
            self.graph_model = KnowledgeGraph

        if self.cognitive_layer_model is None:
            self.cognitive_layer_model = DefaultCognitiveLayer

        if self.intra_layer_score_treshold is None:
            self.intra_layer_score_treshold = config.intra_layer_score_treshold

        if self.embedding_engine is None:
            self.embedding_engine = DefaultEmbeddingEngine()

        if self.connect_documents is None:
            self.connect_documents = config.connect_documents

        if self.chunk_strategy is None:
            self.chunk_strategy = config.chunk_strategy

        if self.chunk_engine is None:
            self.chunk_engine = DefaultChunkEngine()

        if (config_entity is None or config_entity == "llm_engine") and self.llm_engine is None:
            self.llm_engine = OpenAIAdapter(config.openai_key, config.openai_model)

        if (config_entity is None or config_entity == "vector_engine") and self.vector_engine is None:
            try:
                from .databases.vector.weaviate_db import WeaviateAdapter

                if config.weaviate_url is None and config.weaviate_api_key is None:
                    raise EnvironmentError("Weaviate is not configured!")

                self.vector_engine = WeaviateAdapter(
                    config.weaviate_url,
                    config.weaviate_api_key,
                    embedding_engine = self.embedding_engine
                )
            except (EnvironmentError, ModuleNotFoundError):
                if config.qdrant_url is None and config.qdrant_api_key is None:
                    raise EnvironmentError("Qdrant is not configured!")

                self.vector_engine = QDrantAdapter(
                    qdrant_url = config.qdrant_url,
                    qdrant_api_key = config.qdrant_api_key,
                    embedding_engine = self.embedding_engine
                )

        if (config_entity is None or config_entity == "database_directory_path") and self.database_directory_path is None:
            self.database_directory_path = self.system_root_directory + "/" + config.db_path

        if (config_entity is None or config_entity == "database_file_path") and self.database_file_path is None:
            self.database_file_path = self.system_root_directory + "/" + config.db_path + "/" + config.db_name

        if config_entity is not None:
            return getattr(self, config_entity)

        return {
            "llm_engine": self.llm_engine,
            "vector_engine": self.vector_engine,
            "database_engine": self.database_engine,
            "system_root_directory": self.system_root_directory,
            "data_root_directory": self.data_root_directory,
            "graph_engine": self.graph_engine,
            "classification_model": self.classification_model,
            "summarization_model": self.summarization_model,
            "labeling_model": self.labeling_model,
            "graph_model": self.graph_model,
            "cognitive_layer_model": self.cognitive_layer_model,
            "llm_provider": self.llm_provider,
            "intra_layer_score_treshold": self.intra_layer_score_treshold,
            "embedding_engine": self.embedding_engine,
            "connect_documents": self.connect_documents,
            "database_directory_path": self.database_directory_path,
            "database_path": self.database_file_path,
            "chunk_strategy": self.chunk_strategy,
            "chunk_engine": self.chunk_engine,
        }

    def set_config(self, new_config: dict):
        if "system_root_directory" in new_config:
            self.system_root_directory = new_config["system_root_directory"]

        if "data_root_directory" in new_config:
            self.data_root_directory = new_config["data_root_directory"]

        if "database_engine" in new_config:
            self.database_engine = new_config["database_engine"]

        if "vector_engine" in new_config:
            self.vector_engine = new_config["vector_engine"]

        if "llm_engine" in new_config:
            self.llm_engine = new_config["llm_engine"]

        if "graph_engine" in new_config:
            self.graph_engine = new_config["graph_engine"]

        if "classification_model" in new_config:
            self.classification_model = new_config["classification_model"]

        if "summarization_model" in new_config:
            self.summarization_model = new_config["summarization_model"]

        if "labeling_model" in new_config:
            self.labeling_model = new_config["labeling_model"]

        if "graph_model" in new_config:
            self.graph_model = new_config["graph_model"]

        if "llm_provider" in new_config:
            self.llm_provider = new_config["llm_provider"]

        if "cognitive_layer_model" in new_config:
            self.cognitive_layer_model = new_config["cognitive_layer_model"]

        if "intra_layer_score_treshold" in new_config:
            self.intra_layer_score_treshold = new_config["intra_layer_score_treshold"]

        if "embedding_engine" in new_config:
            self.embedding_engine = new_config["embedding_engine"]

        if "connect_documents" in new_config:
            self.connect_documents = new_config["connect_documents"]

        if "chunk_strategy" in new_config:
            self.chunk_strategy = new_config["chunk_strategy"]

        if "chunk_engine" in new_config:
            self.chunk_engine = new_config["chunk_engine"]

infrastructure_config = InfrastructureConfig()



from typing import TypedDict
from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.databases.vector import DataPoint

class TextChunk(TypedDict):
    text: str
    chunk_id: str
    file_metadata: dict

async def add_data_chunks(dataset_data_chunks: dict[str, list[TextChunk]]):
    vector_client = infrastructure_config.get_config("vector_engine")

    identified_chunks = []

    for (dataset_name, chunks) in dataset_data_chunks.items():
        try:
            await vector_client.create_collection(dataset_name)
        except Exception:
            pass

        dataset_chunks = [
            dict(
                chunk_id = chunk["chunk_id"],
                collection = dataset_name,
                text = chunk["text"],
                file_metadata = chunk["file_metadata"],
            ) for chunk in chunks
        ]

        identified_chunks.extend(dataset_chunks)

        await vector_client.create_data_points(
            dataset_name,
            [
                DataPoint(
                    id = chunk["chunk_id"],
                    payload = dict(text = chunk["text"]),
                    embed_field = "text"
                ) for chunk in dataset_chunks
            ],
        )

    return identified_chunks



from typing import Union, Dict

""" Search categories in the graph and return their summary attributes. """

from cognee.shared.data_models import GraphDBType
import networkx as nx

async def search_categories(graph: Union[nx.Graph, any], query_label: str, infrastructure_config: Dict):
    """
    Filter nodes in the graph that contain the specified label and return their summary attributes.
    This function supports both NetworkX graphs and Neo4j graph databases.

    Parameters:
    - graph (Union[nx.Graph, AsyncSession]): The graph object or Neo4j session.
    - query_label (str): The label to filter nodes by.
    - infrastructure_config (Dict): Configuration that includes the graph engine type.

    Returns:
    - Union[Dict, List[Dict]]: For NetworkX, returns a dictionary where keys are node identifiers,
      and values are their 'content_labels' attributes. For Neo4j, returns a list of dictionaries,
      each representing a node with 'nodeId' and 'summary'.
    """
    # Determine which client is in use based on the configuration
    if infrastructure_config.get_config()["graph_engine"] == GraphDBType.NETWORKX:
        # Logic for NetworkX
        return {node: data.get('content_labels') for node, data in graph.nodes(data=True) if query_label in node and 'content_labels' in data}

    elif infrastructure_config.get_config()["graph_engine"] == GraphDBType.NEO4J:
        # Logic for Neo4j
        cypher_query = """
        MATCH (n)
        WHERE $label IN labels(n) AND EXISTS(n.summary)
        RETURN id(n) AS nodeId, n.summary AS summary
        """
        result = await graph.run(cypher_query, label=query_label)
        nodes_summary = [{"nodeId": record["nodeId"], "summary": record["summary"]} for record in await result.list()]
        return nodes_summary

    else:
        raise ValueError("Unsupported graph engine type.")



from typing import List, Union
from os import path
import asyncio
import dlt
import duckdb
import cognee.modules.ingestion as ingestion
from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.files.storage import LocalStorage
from cognee.modules.discovery import discover_directory_datasets

from cognee.utils import send_telemetry



async def add(data_path: Union[str, List[str]], dataset_name: str = None):
    if isinstance(data_path, str):
        # data_path is a data directory path
        if "data://" in data_path:
            return await add_data_directory(data_path.replace("data://", ""), dataset_name)
        # data_path is a file path
        if "file://" in data_path:
            return await add([data_path], dataset_name)
        # data_path is a text
        else:
            file_path = save_text_to_file(data_path, dataset_name)
            return await add([file_path], dataset_name)

    # data_path is a list of file paths or texts
    file_paths = []
    texts = []

    for file_path in data_path:
        if file_path.startswith("/") or file_path.startswith("file://"):
            file_paths.append(file_path)
        else:
            texts.append(file_path)

    awaitables = []

    if len(texts) > 0:
        for text in texts:
            file_paths.append(save_text_to_file(text, dataset_name))

    if len(file_paths) > 0:
        awaitables.append(add_files(file_paths, dataset_name))

    return await asyncio.gather(*awaitables)

async def add_files(file_paths: List[str], dataset_name: str):
    infra_config = infrastructure_config.get_config()
    data_directory_path = infra_config["data_root_directory"]

    LocalStorage.ensure_directory_exists(infra_config["database_directory_path"])

    processed_file_paths = []

    for file_path in file_paths:
        file_path = file_path.replace("file://", "")

        if data_directory_path not in file_path:
            file_name = file_path.split("/")[-1]
            file_directory_path = data_directory_path + "/" + (dataset_name.replace(".", "/") + "/" if dataset_name != "root" else "")
            dataset_file_path = path.join(file_directory_path, file_name)

            LocalStorage.ensure_directory_exists(file_directory_path)

            LocalStorage.copy_file(file_path, dataset_file_path)
            processed_file_paths.append(dataset_file_path)
        else:
            processed_file_paths.append(file_path)

    db = duckdb.connect(infra_config["database_path"])

    destination = dlt.destinations.duckdb(
        credentials = db,
    )

    pipeline = dlt.pipeline(
        pipeline_name = "file_load_from_filesystem",
        destination = destination,
    )

    @dlt.resource(standalone = True, merge_key = "id")
    def data_resources(file_paths: str):
        for file_path in file_paths:
            with open(file_path.replace("file://", ""), mode = "rb") as file:
                classified_data = ingestion.classify(file)

                data_id = ingestion.identify(classified_data)

                file_metadata = classified_data.get_metadata()

                yield {
                    "id": data_id,
                    "name": file_metadata["name"],
                    "file_path": file_metadata["file_path"],
                    "extension": file_metadata["extension"],
                    "mime_type": file_metadata["mime_type"],
                    "keywords": "|".join(file_metadata["keywords"]),
                }

    run_info = pipeline.run(
        data_resources(processed_file_paths),
        table_name = "file_metadata",
        dataset_name = dataset_name.replace(" ", "_").replace(".", "_") if dataset_name is not None else "main_dataset",
        write_disposition = "merge",
    )
    send_telemetry( "COGNEE_ADD")

    return run_info

async def add_data_directory(data_path: str, dataset_name: str = None):
    datasets = discover_directory_datasets(data_path)

    results = []

    for key in datasets.keys():
        if dataset_name is None or key.startswith(dataset_name):
            results.append(add(datasets[key], dataset_name = key))

    return await asyncio.gather(*results)

def save_text_to_file(text: str, dataset_name: str):
    data_directory_path = infrastructure_config.get_config()["data_root_directory"]

    classified_data = ingestion.classify(text)
    data_id = ingestion.identify(classified_data)

    storage_path = data_directory_path + "/" + dataset_name.replace(".", "/")
    LocalStorage.ensure_directory_exists(storage_path)

    text_file_name = data_id + ".txt"
    LocalStorage(storage_path).store(text_file_name, classified_data.get_data())

    return "file://" + storage_path + "/" + text_file_name



""" Chunking strategies for splitting text into smaller parts."""
from __future__ import annotations
import re
from cognee.shared.data_models import ChunkStrategy


class DefaultChunkEngine():
    @staticmethod
    def _split_text_with_regex(
        text: str, separator: str, keep_separator: bool
    ) -> list[str]:
        # Now that we have the separator, split the text
        if separator:
            if keep_separator:
                # The parentheses in the pattern keep the delimiters in the result.
                _splits = re.split(f"({separator})", text)
                splits = [_splits[i] + _splits[i + 1] for i in range(1, len(_splits), 2)]
                if len(_splits) % 2 == 0:
                    splits += _splits[-1:]
                splits = [_splits[0]] + splits
            else:
                splits = re.split(separator, text)
        else:
            splits = list(text)
        return [s for s in splits if s != ""]


    @staticmethod
    def chunk_data(
        chunk_strategy = None,
        source_data = None,
        chunk_size = None,
        chunk_overlap = None,
    ):
        """
        Chunk data based on the specified strategy.

        Parameters:
        - chunk_strategy: The strategy to use for chunking.
        - source_data: The data to be chunked.
        - chunk_size: The size of each chunk.
        - chunk_overlap: The overlap between chunks.

        Returns:
        - The chunked data.
        """

        if chunk_strategy == ChunkStrategy.PARAGRAPH:
            chunked_data = DefaultChunkEngine.chunk_data_by_paragraph(source_data,chunk_size, chunk_overlap)
        elif chunk_strategy == ChunkStrategy.SENTENCE:
            chunked_data = DefaultChunkEngine.chunk_by_sentence(source_data, chunk_size, chunk_overlap)
        elif chunk_strategy == ChunkStrategy.EXACT:
            chunked_data = DefaultChunkEngine.chunk_data_exact(source_data, chunk_size, chunk_overlap)

        return chunked_data


    @staticmethod
    def chunk_data_exact(data_chunks, chunk_size, chunk_overlap):
        data = "".join(data_chunks)
        chunks = []
        for i in range(0, len(data), chunk_size - chunk_overlap):
            chunks.append(data[i:i + chunk_size])
        return chunks


    @staticmethod
    def chunk_by_sentence(data_chunks, chunk_size, overlap):
        # Split by periods, question marks, exclamation marks, and ellipses
        data = "".join(data_chunks)

        # The regular expression is used to find series of charaters that end with one the following chaacters (. ! ? ...)
        sentence_endings = r'(?<=[.!?…]) +'
        sentences = re.split(sentence_endings, data)

        sentence_chunks = []
        for sentence in sentences:
            if len(sentence) > chunk_size:
                chunks = DefaultChunkEngine.chunk_data_exact([sentence], chunk_size, overlap)
                sentence_chunks.extend(chunks)
            else:
                sentence_chunks.append(sentence)
        return sentence_chunks


    @staticmethod
    def chunk_data_by_paragraph(data_chunks, chunk_size, overlap, bound = 0.75):
        data = "".join(data_chunks)
        total_length = len(data)
        chunks = []
        check_bound = int(bound * chunk_size)
        start_idx = 0
        chunk_splitter = "\n\n"

        if data.find("\n\n") == -1:
            chunk_splitter = "\n"

        while start_idx < total_length:
            # Set the end index to the minimum of start_idx + default_chunk_size or total_length
            end_idx = min(start_idx + chunk_size, total_length)

            # Find the next paragraph index within the current chunk and bound
            next_paragraph_index = data.find(chunk_splitter, start_idx + check_bound, end_idx)

            # If a next paragraph index is found within the current chunk
            if next_paragraph_index != -1:
                # Update end_idx to include the paragraph delimiter
                end_idx = next_paragraph_index + 2

            end_index = end_idx + overlap

            chunk_text = data[start_idx:end_index]

            while chunk_text[-1] != "." and end_index < total_length:
                chunk_text += data[end_index]
                end_index += 1

            end_idx = end_index - overlap

            chunks.append(chunk_text.replace("\n", "").strip())

            # Update start_idx to be the current end_idx
            start_idx = end_idx

        return chunks



import logging
from typing import List, Dict
from cognee.infrastructure import infrastructure_config
from.extraction.extract_cognitive_layers import extract_cognitive_layers

logger = logging.getLogger(__name__)

async def get_cognitive_layers(content: str, categories: List[Dict]):
    try:
        return (await extract_cognitive_layers(
            content,
            categories[0],
            infrastructure_config.get_config()["cognitive_layer_model"]
        )).cognitive_layers
    except Exception as error:
        logger.error("Error extracting cognitive layers from content: %s", error, exc_info = True)
        raise error



from cognee.root_dir import get_absolute_path
from .storage.StorageManager import StorageManager
from .storage.LocalStorage import LocalStorage

async def remove_file_from_storage(file_path: str):
    storage_manager = StorageManager(LocalStorage(get_absolute_path("data/files")))

    storage_manager.remove(file_path)



from typing import BinaryIO
from cognee.root_dir import get_absolute_path
from .storage.StorageManager import StorageManager
from .storage.LocalStorage import LocalStorage

async def add_file_to_storage(file_path: str, file: BinaryIO):
    storage_manager = StorageManager(LocalStorage(get_absolute_path("data/files")))

    storage_manager.store(file_path, file)



def is_text_content(content):
    # Check for null bytes
    if b'\0' in content:
        return False

    # Check for common text encodings (BOMs)
    if content.startswith((b'\xEF\xBB\xBF',  # UTF-8
                            b'\xFF\xFE',      # UTF-16 LE
                            b'\xFE\xFF',      # UTF-16 BE
                            b'\x00\x00\xFE\xFF',  # UTF-32 LE
                            b'\xFF\xFE\x00\x00',  # UTF-32 BE
                            )):
        return True

    # Check for ASCII characters
    if all(0x20 <= byte <= 0x7E or byte in (b'\n', b'\r', b'\t') for byte in content):
        return True

    # Check for common line break characters
    if b'\n' in content or b'\r' in content:
        return True

    # If no obvious indicators found, assume it's a text file
    return True



from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.infrastructure.llm.prompts import render_prompt

async def extract_content_graph(content: str, cognitive_layer: str, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    system_prompt = render_prompt("generate_graph_prompt.txt", { "layer": cognitive_layer })
    output = await llm_client.acreate_structured_output(content, system_prompt, response_model)

    return output.model_dump()



from typing import Union
from pydantic import BaseModel

class DataPoint(BaseModel):
    id: str
    payload: dict[str, Union[str, dict[str, str]]]
    embed_field: str = "value"

    def get_embeddable_data(self):
        return self.payload[self.embed_field]



from datetime import datetime
from cognee.shared.data_models import DefaultGraphModel, Relationship, UserProperties, UserLocation
from cognee.modules.cognify.graph.create import create_semantic_graph

async def initialize_graph(root_id: str, graphdatamodel, graph_client):
    if graphdatamodel:
        graph = graphdatamodel(id = root_id)
        graph_ = await create_semantic_graph(graph, graph_client)
        return graph_
    else:
        print("Creating default graph")

        graph = DefaultGraphModel(
            node_id = root_id,
            user_properties = UserProperties(
                custom_properties = { "age": "30" },
                location = UserLocation(
                    location_id = "ny",
                    description = "New York",
                    default_relationship = Relationship(
                        type = "located_in",
                        source = "UserProperties",
                        target = "ny",
                    )
                ),
                default_relationship = Relationship(
                    type = "has_properties",
                    source = root_id,
                    target = "UserProperties",
                )
            ),
            default_relationship = Relationship(
                type = "has_properties",
                source = root_id,
                target = "UserProperties"
            ),
            default_fields = {
                "created_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                "updated_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            }
        )

        return await create_semantic_graph(graph, graph_client)



""" This module contains the function to find the neighbours of a given node in the graph"""


from typing import Union, Dict
import networkx as nx
from cognee.shared.data_models import GraphDBType
async def search_adjacent(graph: Union[nx.Graph, any], query: str, infrastructure_config: Dict, other_param: dict = None) -> Dict[str, str]:
    """
    Find the neighbours of a given node in the graph and return their descriptions.
    Supports both NetworkX graphs and Neo4j graph databases based on the configuration.

    Parameters:
    - graph (Union[nx.Graph, AsyncSession]): The graph object or Neo4j session.
    - query (str): Unused in this implementation but could be used for future enhancements.
    - infrastructure_config (Dict): Configuration that includes the graph engine type.
    - other_param (dict, optional): A dictionary that may contain 'node_id' to specify the node.

    Returns:
    - Dict[str, str]: A dictionary containing the unique identifiers and descriptions of the neighbours of the given node.
    """
    node_id = other_param.get('node_id') if other_param else None

    if node_id is None:
        return {}

    if infrastructure_config.get_config()["graph_engine"] == GraphDBType.NETWORKX:
        if node_id not in graph:
            return {}

        neighbors = list(graph.neighbors(node_id))
        neighbor_descriptions = {neighbor: graph.nodes[neighbor].get('description') for neighbor in neighbors}
        return neighbor_descriptions

    elif infrastructure_config.get_config()["graph_engine"] == GraphDBType.NEO4J:
        cypher_query = """
        MATCH (node {id: $node_id})-[:CONNECTED_TO]->(neighbor)
        RETURN neighbor.id AS neighbor_id, neighbor.description AS description
        """
        results = await graph.run(cypher_query, node_id=node_id)
        neighbor_descriptions = {record["neighbor_id"]: record["description"] for record in await results.list() if "description" in record}
        return neighbor_descriptions

    else:
        raise ValueError("Unsupported graph engine type in the configuration.")


from typing import List, Protocol

class EmbeddingEngine(Protocol):
    async def embed_text(self, text: str) -> List[float]:
        raise NotImplementedError()

    def get_vector_size(self) -> int:
        raise NotImplementedError()



import asyncio
from uuid import UUID, uuid4
from cognee.infrastructure.data import Data, Dataset
from .add_data_to_dataset import add_data_to_dataset
from .data_types import IngestionData

async def save(dataset_id: UUID, dataset_name: str, data_id: UUID, data: IngestionData):
    file_path = uuid4().hex + "." + data.get_extension()

    promises = []

    promises.append(
        add_data_to_dataset(
            Dataset(
                id = dataset_id,
                name = dataset_name if dataset_name else dataset_id.hex
            ),
            Data(
                id = data_id,
                raw_data_location = file_path,
                name = data.metadata["name"],
                meta_data = data.metadata
            )
        )
    )

    await asyncio.gather(*promises)



from io import BufferedReader
from typing import Union, BinaryIO
from .exceptions import IngestionException
from .data_types import create_text_data, create_binary_data

def classify(data: Union[str, BinaryIO]):
    if isinstance(data, str):
        return create_text_data(data)

    if isinstance(data, BufferedReader):
        return create_binary_data(data)

    raise IngestionException(f"Type of data sent to cognee.add(data_path: string | List[string]) not supported: {type(data)}")



import asyncio
from typing import List, Type
from pydantic import BaseModel
import instructor
from tenacity import retry, stop_after_attempt
from openai import AsyncOpenAI
import openai
from cognee.infrastructure.llm.llm_interface import LLMInterface
from cognee.infrastructure.llm.prompts import read_query_prompt


class GenericAPIAdapter(LLMInterface):
    """Adapter for Ollama's API"""

    def __init__(self, api_endpoint, api_key: str, model: str):
        self.aclient =  instructor.patch(
            AsyncOpenAI(
                base_url = api_endpoint,
                api_key = api_key,  # required, but unused
            ),
            mode = instructor.Mode.JSON,
        )
        self.model = model

    @retry(stop = stop_after_attempt(5))
    def completions_with_backoff(self, **kwargs):
        """Wrapper around ChatCompletion.create w/ backoff"""
        # Local model
        return openai.chat.completions.create(**kwargs)

    @retry(stop = stop_after_attempt(5))
    async def acompletions_with_backoff(self, **kwargs):
        """Wrapper around ChatCompletion.acreate w/ backoff"""
        return await openai.chat.completions.acreate(**kwargs)

    @retry(stop = stop_after_attempt(5))
    async def acreate_embedding_with_backoff(self, input: List[str], model: str = "text-embedding-3-large"):
        """Wrapper around Embedding.acreate w/ backoff"""

        return await self.aclient.embeddings.create(input = input, model = model)

    async def async_get_embedding_with_backoff(self, text, model="text-embedding-3-large"):
        """To get text embeddings, import/call this function
        It specifies defaults + handles rate-limiting + is async"""
        text = text.replace("\n", " ")
        response = await self.aclient.embeddings.create(input = text, model = model)
        embedding = response.data[0].embedding
        return embedding

    @retry(stop = stop_after_attempt(5))
    def create_embedding_with_backoff(self, **kwargs):
        """Wrapper around Embedding.create w/ backoff"""
        return openai.embeddings.create(**kwargs)

    def get_embedding_with_backoff(self, text: str, model: str = "text-embedding-3-large"):
        """To get text embeddings, import/call this function
        It specifies defaults + handles rate-limiting
        :param text: str
        :param model: str
        """
        text = text.replace("\n", " ")
        response = self.create_embedding_with_backoff(input=[text], model=model)
        embedding = response.data[0].embedding
        return embedding

    async def async_get_batch_embeddings_with_backoff(self, texts: List[str], models: List[str]):
        """To get multiple text embeddings in parallel, import/call this function
        It specifies defaults + handles rate-limiting + is async"""
        # Collect all coroutines
        coroutines = (self.async_get_embedding_with_backoff(text, model)
                      for text, model in zip(texts, models))

        # Run the coroutines in parallel and gather the results
        embeddings = await asyncio.gather(*coroutines)

        return embeddings

    @retry(stop=stop_after_attempt(5))
    async def acreate_structured_output(self, text_input: str, system_prompt: str,
                                        response_model: Type[BaseModel]) -> BaseModel:
        """Generate a response from a user query."""
        return await self.aclient.chat.completions.create(
            model=self.model,
            messages=[
                {
                    "role": "user",
                    "content": f"""Use the given format to
                    extract information from the following input: {text_input}. {system_prompt} """,
                }
            ],
            response_model=response_model,
        )

    def show_prompt(self, text_input: str, system_prompt: str) -> str:
        """Format and display the prompt for a user query."""
        if not text_input:
            text_input = "No user input provided."
        if not system_prompt:
            raise ValueError("No system prompt path provided.")
        system_prompt = read_query_prompt(system_prompt)

        formatted_prompt = f"""System Prompt:\n{system_prompt}\n\nUser Input:\n{text_input}\n""" if system_prompt else None
        return formatted_prompt



"""Here we update the graph with content summary that summarizer produced"""

async def add_summary_nodes(graph_client, document_id, summary):
    summary_node_id = f"DATA_SUMMARY__{document_id}"

    await graph_client.add_node(
        summary_node_id,
        dict(
            name = "Summary",
            summary = summary["summary"],
        ),
    )

    await graph_client.add_edge(document_id, summary_node_id, relationship_name = "summarized_as")


    description_node_id = f"DATA_DESCRIPTION__{document_id}"

    await graph_client.add_node(
        description_node_id,
        dict(
            name = "Description",
            summary = summary["description"],
        ),
    )

    await graph_client.add_edge(document_id, description_node_id, relationship_name = "described_as")



from typing import BinaryIO, TypedDict
from cognee.infrastructure.data.utils.extract_keywords import extract_keywords
from .extract_text_from_file import extract_text_from_file
from .guess_file_type import guess_file_type


class FileMetadata(TypedDict):
    name: str
    mime_type: str
    extension: str
    keywords: list[str]

def get_file_metadata(file: BinaryIO) -> FileMetadata:
    file.seek(0)
    file_type = guess_file_type(file)

    file.seek(0)
    file_text = extract_text_from_file(file, file_type)
    keywords = extract_keywords(file_text)

    file_path = file.name
    file_name = file_path.split("/")[-1].split(".")[0]

    return FileMetadata(
        name = file_name,
        file_path = file_path,
        mime_type = file_type.mime,
        extension = file_type.extension,
        keywords = keywords
    )



from dsp.utils import deduplicate
from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client


async def search_similarity(query: str, graph):
    graph_db_type = infrastructure_config.get_config()["graph_engine"]

    graph_client = await get_graph_client(graph_db_type)

    layer_nodes = await graph_client.get_layer_nodes()

    unique_layer_uuids = set(node["layer_id"] for node in layer_nodes)

    graph_nodes = []

    for layer_id in unique_layer_uuids:
        vector_engine = infrastructure_config.get_config()["vector_engine"]

        results = await vector_engine.search(layer_id, query_text = query, limit = 10)

        if len(results) > 0:
            graph_nodes.extend([
                dict(
                    layer_id = result.payload["references"]["cognitive_layer"],
                    node_id = result.payload["references"]["node_id"],
                    score = result.score,
                ) for result in results if result.score > 0.5
            ])

    if len(graph_nodes) == 0:
        return []

    relevant_context = []

    for graph_node_data in graph_nodes:
        graph_node = await graph_client.extract_node(graph_node_data["node_id"])

        if "chunk_collection" not in graph_node and "chunk_id" not in graph_node:
            continue

        vector_point = await vector_engine.retrieve(graph_node["chunk_collection"], graph_node["chunk_id"])

        relevant_context.append(vector_point.payload["text"])

    return deduplicate(relevant_context)



"""Data models for the cognitive architecture."""

from enum import Enum, auto
from typing import Optional, List, Union, Dict, Any
from pydantic import BaseModel, Field

class Node(BaseModel):
    """Node in a knowledge graph."""
    id: str
    entity_name: str
    entity_type: str
    entity_description: str

class Edge(BaseModel):
    """Edge in a knowledge graph."""
    source_node_id: str
    target_node_id: str
    relationship_name: str

class KnowledgeGraph(BaseModel):
    """Knowledge graph."""
    nodes: List[Node] = Field(..., default_factory=list)
    edges: List[Edge] = Field(..., default_factory=list)

class GraphQLQuery(BaseModel):
    """GraphQL query."""
    query: str



class Answer(BaseModel):
    """Answer."""
    answer: str

class ChunkStrategy(Enum):
    EXACT = "exact"
    PARAGRAPH = "paragraph"
    SENTENCE = "sentence"

class MemorySummary(BaseModel):
    """ Memory summary. """
    nodes: List[Node] = Field(..., default_factory=list)
    edges: List[Edge] = Field(..., default_factory=list)


class TextSubclass(str, Enum):
    ARTICLES = "Articles, essays, and reports"
    BOOKS = "Books and manuscripts"
    NEWS_STORIES = "News stories and blog posts"
    RESEARCH_PAPERS = "Research papers and academic publications"
    SOCIAL_MEDIA = "Social media posts and comments"
    WEBSITE_CONTENT = "Website content and product descriptions"
    PERSONAL_NARRATIVES = "Personal narratives and stories"
    SPREADSHEETS = "Spreadsheets and tables"
    FORMS = "Forms and surveys"
    DATABASES = "Databases and CSV files"
    SOURCE_CODE = "Source code in various programming languages"
    SHELL_SCRIPTS = "Shell commands and scripts"
    MARKUP_LANGUAGES = "Markup languages (HTML, XML)"
    STYLESHEETS = "Stylesheets (CSS) and configuration files (YAML, JSON, INI)"
    CHAT_TRANSCRIPTS = "Chat transcripts and messaging history"
    CUSTOMER_SERVICE_LOGS = "Customer service logs and interactions"
    CONVERSATIONAL_AI = "Conversational AI training data"
    TEXTBOOK_CONTENT = "Textbook content and lecture notes"
    EXAM_QUESTIONS = "Exam questions and academic exercises"
    E_LEARNING_MATERIALS = "E-learning course materials"
    POETRY = "Poetry and prose"
    SCRIPTS = "Scripts for plays, movies, and television"
    SONG_LYRICS = "Song lyrics"
    MANUALS = "Manuals and user guides"
    TECH_SPECS = "Technical specifications and API documentation"
    HELPDESK_ARTICLES = "Helpdesk articles and FAQs"
    LEGAL_CONTRACTS = "Contracts and agreements"
    LAWS = "Laws, regulations, and legal case documents"
    POLICY_DOCUMENTS = "Policy documents and compliance materials"
    CLINICAL_TRIALS = "Clinical trial reports"
    PATIENT_RECORDS = "Patient records and case notes"
    SCIENTIFIC_ARTICLES = "Scientific journal articles"
    FINANCIAL_REPORTS = "Financial reports and statements"
    BUSINESS_PLANS = "Business plans and proposals"
    MARKET_RESEARCH = "Market research and analysis reports"
    AD_COPIES = "Ad copies and marketing slogans"
    PRODUCT_CATALOGS = "Product catalogs and brochures"
    PRESS_RELEASES = "Press releases and promotional content"
    PROFESSIONAL_EMAILS = "Professional and formal correspondence"
    PERSONAL_EMAILS = "Personal emails and letters"
    IMAGE_CAPTIONS = "Image and video captions"
    ANNOTATIONS = "Annotations and metadata for various media"
    VOCAB_LISTS = "Vocabulary lists and grammar rules"
    LANGUAGE_EXERCISES = "Language exercises and quizzes"
    LEGAL_AND_REGULATORY_DOCUMENTS = "Legal and Regulatory Documents"
    OTHER_TEXT = "Other types of text data"

class AudioSubclass(str, Enum):
    MUSIC_TRACKS = "Music tracks and albums"
    PODCASTS = "Podcasts and radio broadcasts"
    AUDIOBOOKS = "Audiobooks and audio guides"
    INTERVIEWS = "Recorded interviews and speeches"
    SOUND_EFFECTS = "Sound effects and ambient sounds"
    OTHER_AUDIO = "Other types of audio recordings"

class ImageSubclass(str, Enum):
    PHOTOGRAPHS = "Photographs and digital images"
    ILLUSTRATIONS = "Illustrations, diagrams, and charts"
    INFOGRAPHICS = "Infographics and visual data representations"
    ARTWORK = "Artwork and paintings"
    SCREENSHOTS = "Screenshots and graphical user interfaces"
    OTHER_IMAGES = "Other types of images"


class VideoSubclass(str, Enum):
    MOVIES = "Movies and short films"
    DOCUMENTARIES = "Documentaries and educational videos"
    TUTORIALS = "Video tutorials and how-to guides"
    ANIMATED_FEATURES = "Animated features and cartoons"
    LIVE_EVENTS = "Live event recordings and sports broadcasts"
    OTHER_VIDEOS = "Other types of video content"

class MultimediaSubclass(str, Enum):
    WEB_CONTENT = "Interactive web content and games"
    VR_EXPERIENCES = "Virtual reality (VR) and augmented reality (AR) experiences"
    MIXED_MEDIA = "Mixed media presentations and slide decks"
    E_LEARNING_MODULES = "E-learning modules with integrated multimedia"
    DIGITAL_EXHIBITIONS = "Digital exhibitions and virtual tours"
    OTHER_MULTIMEDIA = "Other types of multimedia content"

class Model3DSubclass(str, Enum):
    ARCHITECTURAL_RENDERINGS = "Architectural renderings and building plans"
    PRODUCT_MODELS = "Product design models and prototypes"
    ANIMATIONS = "3D animations and character models"
    SCIENTIFIC_VISUALIZATIONS = "Scientific simulations and visualizations"
    VR_OBJECTS = "Virtual objects for AR/VR applications"
    OTHER_3D_MODELS = "Other types of 3D models"

class ProceduralSubclass(str, Enum):
    TUTORIALS_GUIDES = "Tutorials and step-by-step guides"
    WORKFLOW_DESCRIPTIONS = "Workflow and process descriptions"
    SIMULATIONS = "Simulation and training exercises"
    RECIPES = "Recipes and crafting instructions"
    OTHER_PROCEDURAL = "Other types of procedural content"

class ContentType(BaseModel):
    """Base class for different types of content."""
    type: str

class TextContent(ContentType):
    type: str = "TEXT"
    subclass: List[TextSubclass]

class AudioContent(ContentType):
    type: str = "AUDIO"
    subclass: List[AudioSubclass]

class ImageContent(ContentType):
    type: str = "IMAGE"
    subclass: List[ImageSubclass]

class VideoContent(ContentType):
    type: str = "VIDEO"
    subclass: List[VideoSubclass]

class MultimediaContent(ContentType):
    type: str = "MULTIMEDIA"
    subclass: List[MultimediaSubclass]

class Model3DContent(ContentType):
    type: str = "3D_MODEL"
    subclass: List[Model3DSubclass]

class ProceduralContent(ContentType):
    type: str = "PROCEDURAL"
    subclass: List[ProceduralSubclass]

class DefaultContentPrediction(BaseModel):
    """Class for a single class label prediction."""
    label: Union[
        TextContent,
        AudioContent,
        ImageContent,
        VideoContent,
        MultimediaContent,
        Model3DContent,
        ProceduralContent,
    ]


class SummarizedContent(BaseModel):
    """Class for a single class label summary and description."""
    summary: str
    description: str

class LabeledContent(BaseModel):
    """Class for a single class label summary."""
    content_labels: str



class CognitiveLayerSubgroup(BaseModel):
    """ CognitiveLayerSubgroup in a general layer """
    id: int
    name: str
    description: str


class DefaultCognitiveLayer(BaseModel):
    """Cognitive  layer"""
    category_name: str
    cognitive_layers: List[CognitiveLayerSubgroup] = Field(..., default_factory=list)


class GraphDBType(Enum):
    NETWORKX = auto()
    NEO4J = auto()


# Models for representing different entities
class Relationship(BaseModel):
    type: str
    source: Optional[str] = None
    target: Optional[str] = None
    properties: Optional[Dict[str, Any]] = None

class DocumentType(BaseModel):
    type_id: str
    description: str
    default_relationship: Relationship = Relationship(type = "is_type")

class Category(BaseModel):
    category_id: str
    name: str
    default_relationship: Relationship = Relationship(type = "categorized_as")

class Document(BaseModel):
    id: str
    title: str
    description: Optional[str] = None

class UserLocation(BaseModel):
    location_id: str
    description: str
    default_relationship: Relationship = Relationship(type = "located_in")

class UserProperties(BaseModel):
    custom_properties: Optional[Dict[str, Any]] = None
    location: Optional[UserLocation] = None

class DefaultGraphModel(BaseModel):
    node_id: str
    user_properties: UserProperties = UserProperties()
    documents: List[Document] = []
    default_fields: Optional[Dict[str, Any]] = {}
    default_relationship: Relationship = Relationship(type = "has_properties")



"""Factory function to get the appropriate graph client based on the graph type."""

from cognee.config import Config
from cognee.shared.data_models import GraphDBType
from cognee.infrastructure import infrastructure_config
from .graph_db_interface import GraphDBInterface
from .networkx.adapter import NetworkXAdapter

config = Config()
config.load()


async def get_graph_client(graph_type: GraphDBType, graph_file_name: str = None) -> GraphDBInterface :
    """Factory function to get the appropriate graph client based on the graph type."""
    graph_file_path = f"{infrastructure_config.get_config('database_directory_path')}/{graph_file_name if graph_file_name else config.graph_filename}"

    if graph_type == GraphDBType.NEO4J:
        try:
            from .neo4j_driver.adapter import Neo4jAdapter

            return Neo4jAdapter(
                graph_database_url = config.graph_database_url,
                graph_database_username = config.graph_database_username,
                graph_database_password = config.graph_database_password
            )
        except:
            pass
            
    graph_client = NetworkXAdapter(filename = graph_file_path)
    await graph_client.load_graph_from_file()
    return graph_client



class PineconeVectorDB(VectorDB):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_pinecone(self.index_name)

    def init_pinecone(self, index_name):
        # Pinecone initialization logic
        pass



from typing import List
from datetime import datetime, timezone
from sqlalchemy.orm import relationship, Mapped
from sqlalchemy import Column, Text, DateTime, UUID
from cognee.infrastructure.databases.relational import ModelBase
from .DatasetData import DatasetData

class Dataset(ModelBase):
    __tablename__ = "dataset"

    id = Column(UUID, primary_key = True)
    name = Column(Text)
    description = Column(Text, nullable = True)

    created_at = Column(DateTime, default = datetime.now(timezone.utc))
    updated_at = Column(DateTime, onupdate = datetime.now(timezone.utc))

    data: Mapped[List["Data"]] = relationship(
        secondary = DatasetData.__tablename__,
        back_populates = "datasets"
    )



"""Adapter for NetworkX graph database."""

import os
import json
import logging
from typing import Dict, Any, List
import aiofiles
import aiofiles.os as aiofiles_os
import networkx as nx
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface

logger = logging.getLogger("NetworkXAdapter")

class NetworkXAdapter(GraphDBInterface):
    _instance = None  # Class variable to store the singleton instance

    def __new__(cls, filename):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.filename = filename
        return cls._instance

    def __init__(self, filename = "cognee_graph.pkl"):
        self.filename = filename
        self.graph = nx.MultiDiGraph()

    async def graph(self):
        return self.graph

    async def add_node(
        self,
        node_id: str,
        node_properties,
    ) -> None:
        if not self.graph.has_node(id):
            self.graph.add_node(node_id, **node_properties)
            await self.save_graph_to_file(self.filename)

    async def add_nodes(
        self,
        nodes: List[tuple[str, dict]],
    ) -> None:
        self.graph.add_nodes_from(nodes)
        await self.save_graph_to_file(self.filename)
    
    async def add_edge(
        self,
        from_node: str,
        to_node: str,
        relationship_name: str,
        edge_properties: Dict[str, Any] = None,
    ) -> None:
        self.graph.add_edge(from_node, to_node, key = relationship_name, **(edge_properties if edge_properties else {}))
        await self.save_graph_to_file(self.filename)

    async def add_edges(
        self,
        edges: tuple[str, str, str, dict],
    ) -> None:
        self.graph.add_edges_from(edges)
        await self.save_graph_to_file(self.filename)

    async def delete_node(self, node_id: str) -> None:
        """Asynchronously delete a node from the graph if it exists."""
        if self.graph.has_node(id):
            self.graph.remove_node(id)
            await self.save_graph_to_file(self.filename)

    async def extract_node_description(self, node_id: str) -> Dict[str, Any]:
        descriptions = []

        if self.graph.has_node(node_id):
            # Get the attributes of the node
            for neighbor in self.graph.neighbors(node_id):
                # Get the attributes of the neighboring node
                attributes = self.graph.nodes[neighbor]

                # Ensure all required attributes are present before extracting description
                if all(key in attributes for key in ["id", "layer_id", "description"]):
                    descriptions.append({
                        "id": attributes["id"],
                        "layer_id": attributes["layer_id"],
                        "description": attributes["description"],
                    })

        return descriptions

    async def get_layer_nodes(self):
        layer_nodes = []

        for _, data in self.graph.nodes(data = True):
            if "layer_id" in data:
                layer_nodes.append(data)

        return layer_nodes

    async def extract_node(self, node_id: str) -> dict:
        if self.graph.has_node(node_id):
            return self.graph.nodes[node_id]

        return None


    async def save_graph_to_file(self, file_path: str=None) -> None:
        """Asynchronously save the graph to a file in JSON format."""
        if not file_path:
            file_path = self.filename

        graph_data = nx.readwrite.json_graph.node_link_data(self.graph)

        async with aiofiles.open(file_path, "w") as file:
            await file.write(json.dumps(graph_data))

    async def load_graph_from_file(self, file_path: str = None):
        """Asynchronously load the graph from a file in JSON format."""
        if not file_path:
            file_path = self.filename
        try:
            if os.path.exists(file_path):
                async with aiofiles.open(file_path, "r") as file:
                    graph_data = json.loads(await file.read())
                    self.graph = nx.readwrite.json_graph.node_link_graph(graph_data)
                    return self.graph
            else:
                # Log that the file does not exist and an empty graph is initialized
                logger.warning("File %s not found. Initializing an empty graph.", file_path)
                self.graph = nx.MultiDiGraph()  # Use MultiDiGraph to keep it consistent with __init__
                return self.graph
        except Exception as error:
            logger.error("Failed to load graph from file: %s", file_path)
            # Initialize an empty graph in case of error
            self.graph = nx.MultiDiGraph()
            return self.graph

    async def delete_graph(self, path: str = None):
        """Asynchronously delete the graph file from the filesystem."""
        if path is None:
            path = self.filename  # Assuming self.filename is defined elsewhere and holds the default graph file path
        try:
            await aiofiles_os.remove(path)
            logger.info("Graph deleted successfully.")
        except Exception as error:
            logger.error("Failed to delete graph: %s", error)



from typing import List

import instructor
from openai import AsyncOpenAI
from fastembed import TextEmbedding
from fastembed import TextEmbedding
from openai import AsyncOpenAI

from cognee.config import Config
from cognee.root_dir import get_absolute_path
from .EmbeddingEngine import EmbeddingEngine

config = Config()
config.load()

class DefaultEmbeddingEngine(EmbeddingEngine):
    async def embed_text(self, text: List[str]) -> List[float]:
        embedding_model = TextEmbedding(model_name = config.embedding_model, cache_dir = get_absolute_path("cache/embeddings"))
        embeddings_list = list(map(lambda embedding: embedding.tolist(), embedding_model.embed(text)))

        return embeddings_list

    def get_vector_size(self) -> int:
        return config.embedding_dimensions


class OpenAIEmbeddingEngine(EmbeddingEngine):
    async def embed_text(self, text: List[str]) -> List[float]:

        OPENAI_API_KEY = config.openai_key

        aclient = instructor.patch(AsyncOpenAI())
        text = text.replace("\n", " ")
        response = await aclient.embeddings.create(input = text, model = config.openai_embedding_model)
        embedding = response.data[0].embedding
        # embeddings_list = list(map(lambda embedding: embedding.tolist(), embedding_model.embed(text)))
        return embedding


    def get_vector_size(self) -> int:
        return config.openai_embedding_dimensions




import logging
from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.data import Dataset, Data
from cognee.infrastructure.files import remove_file_from_storage
from cognee.infrastructure.databases.relational import DatabaseEngine

logger = logging.getLogger(__name__)

async def add_data_to_dataset(dataset: Dataset, data: Data):
    db_engine: DatabaseEngine = infrastructure_config.get_config()["database_engine"]

    existing_dataset = (await db_engine.query_entity(dataset)).scalar()
    existing_data = (await db_engine.query_entity(data)).scalar()

    if existing_dataset:
        if existing_data:
            await remove_old_raw_data(existing_data.raw_data_location)

            def update_raw_data():
                existing_data.raw_data_location = data.raw_data_location

            await db_engine.update(update_raw_data)

            if existing_dataset.id == dataset.id and dataset.name is not None:
                def update_name(): existing_dataset.name = dataset.name
                await db_engine.update(update_name)
        else:
            await db_engine.update(lambda: existing_dataset.data.append(data))
    else:
        if existing_data:
            await remove_old_raw_data(existing_data.raw_data_location)

            existing_data.raw_data_location = data.raw_data_location

            await db_engine.update(lambda: dataset.data.append(existing_data))
        else:
            await db_engine.update(lambda: dataset.data.append(data))

        await db_engine.create(dataset)

async def remove_old_raw_data(data_location: str):
    try:
        await remove_file_from_storage(data_location)
    except Exception:
        logger.error("Failed to remove old raw data file: %s", data_location)



from typing import List, Protocol, Optional
from abc import abstractmethod
from .models.DataPoint import DataPoint

class VectorDBInterface(Protocol):
    """ Collections """
    @abstractmethod
    async def create_collection(
        self,
        collection_name: str
    ): raise NotImplementedError

    # @abstractmethod
    # async def update_collection(
    #     self,
    #     collection_name: str,
    #     collection_config: object
    # ): raise NotImplementedError

    # @abstractmethod
    # async def delete_collection(
    #     self,
    #     collection_name: str
    # ): raise NotImplementedError

    # @abstractmethod
    # async def create_vector_index(
    #     self,
    #     collection_name: str,
    #     vector_index_config: object
    # ): raise NotImplementedError

    # @abstractmethod
    # async def create_data_index(
    #     self,
    #     collection_name: str,
    #     vector_index_config: object
    # ): raise NotImplementedError

    """ Data points """
    @abstractmethod
    async def create_data_points(
        self,
        collection_name: str,
        data_points: List[DataPoint]
    ): raise NotImplementedError

    # @abstractmethod
    # async def get_data_point(
    #     self,
    #     collection_name: str,
    #     data_point_id: str
    # ): raise NotImplementedError

    # @abstractmethod
    # async def update_data_point(
    #     self,
    #     collection_name: str,
    #     data_point_id: str,
    #     payload: object
    # ): raise NotImplementedError

    # @abstractmethod
    # async def delete_data_point(
    #     self,
    #     collection_name: str,
    #     data_point_id: str
    # ): raise NotImplementedError

    """ Search """
    @abstractmethod
    async def search(
        self,
        collection_name: str,
        query_text: Optional[str],
        query_vector: Optional[List[float]],
        limit: int,
        with_vector: bool = False

    ): raise NotImplementedError

    @abstractmethod
    async def batch_search(
        self,
        collection_name: str,
        query_texts: List[str],
        limit: int,
        with_vectors: bool = False
    ): raise NotImplementedError



from os import path, listdir

def discover_directory_datasets(root_dir_path: str, parent_dir: str = None):
    datasets = {}

    for file_or_dir in listdir(root_dir_path):
        if path.isdir(path.join(root_dir_path, file_or_dir)):
            dataset_name = file_or_dir if parent_dir is None else f"{parent_dir}.{file_or_dir}"

            nested_datasets = discover_directory_datasets(path.join(root_dir_path, file_or_dir), dataset_name)

            for dataset in nested_datasets.keys():
                datasets[dataset] = nested_datasets[dataset]
        else:
            if parent_dir not in datasets:
                datasets[parent_dir] = []

            datasets[parent_dir].append(path.join(root_dir_path, file_or_dir))

    return datasets



from uuid import uuid4
from datetime import datetime, timezone
from sqlalchemy import Column, DateTime, UUID, ForeignKey
from cognee.infrastructure.databases.relational import ModelBase

class DatasetData(ModelBase):
    __tablename__ = "dataset_data"

    id = Column(UUID, primary_key = True, default = uuid4())

    created_at = Column(DateTime, default = datetime.now(timezone.utc))

    dataset_id = Column("dataset", UUID, ForeignKey("dataset.id"), primary_key = True)
    data_id = Column("data", UUID, ForeignKey("data.id"), primary_key = True)



"""Configuration for cognee - cognitive architecture framework."""
import logging
import os
import configparser
import uuid
from typing import Optional, Dict, Any
from dataclasses import dataclass, field
from pathlib import Path
from dotenv import load_dotenv
from cognee.root_dir import get_absolute_path
from cognee.shared.data_models import ChunkStrategy

base_dir = Path(__file__).resolve().parent.parent
# Load the .env file from the base directory
dotenv_path = base_dir / ".env"
load_dotenv(dotenv_path=dotenv_path)


@dataclass
class Config:
    """ Configuration for cognee - cognitive architecture framework. """
    cognee_dir: str = field(
        default_factory=lambda: os.getenv("COG_ARCH_DIR", "cognee")
    )
    config_path: str = field(
        default_factory=lambda: os.path.join(
            os.getenv("COG_ARCH_DIR", "cognee"), "config"
        )
    )


    system_root_directory = get_absolute_path(".cognee_system")
    data_root_directory = os.getenv("DATA_PATH", get_absolute_path(".data"))

    vectordb: str = os.getenv("VECTORDB", "weaviate")

    qdrant_path: str = os.getenv("QDRANT_PATH")
    qdrant_url: str = os.getenv("QDRANT_URL")
    qdrant_api_key: str = os.getenv("QDRANT_API_KEY")

    db_path = str = os.getenv("COGNEE_DB_PATH", "databases")
    db_name: str = os.getenv("DB_NAME", "cognee.db")
    db_host: str = os.getenv("DB_HOST", "localhost")
    db_port: str = os.getenv("DB_PORT", "5432")
    db_user: str = os.getenv("DB_USER", "cognee")
    db_password: str = os.getenv("DB_PASSWORD", "cognee")

    sqlalchemy_logging: bool = os.getenv("SQLALCHEMY_LOGGING", True)

    graph_filename = os.getenv("GRAPH_NAME", "cognee_graph.pkl")

    # Model parameters
    llm_provider: str = os.getenv("LLM_PROVIDER","openai") #openai, or custom or ollama
    custom_model: str = os.getenv("CUSTOM_LLM_MODEL", "mistralai/Mixtral-8x7B-Instruct-v0.1") #"mistralai/Mixtral-8x7B-Instruct-v0.1"
    custom_endpoint: str = os.getenv("CUSTOM_ENDPOINT", "https://api.endpoints.anyscale.com/v1") #"https://api.endpoints.anyscale.com/v1" # pass claude endpoint
    custom_key: Optional[str] = os.getenv("CUSTOM_LLM_API_KEY")
    ollama_endpoint: str = os.getenv("CUSTOM_OLLAMA_ENDPOINT", "http://localhost:11434/v1") #"http://localhost:11434/v1"
    ollama_key: Optional[str] = "ollama"
    ollama_model: str = os.getenv("CUSTOM_OLLAMA_MODEL", "mistral:instruct") #"mistral:instruct"
    openai_model: str = os.getenv("OPENAI_MODEL", "gpt-4-1106-preview" ) #"gpt-4-1106-preview"
    model_endpoint: str = "openai"
    openai_key: Optional[str] = os.getenv("OPENAI_API_KEY")
    openai_temperature: float = float(os.getenv("OPENAI_TEMPERATURE", 0.0))
    openai_embedding_model = "text-embedding-3-large"
    openai_embedding_dimensions = 3072

    graphistry_username = os.getenv("GRAPHISTRY_USERNAME")
    graphistry_password = os.getenv("GRAPHISTRY_PASSWORD")

    # Embedding parameters
    embedding_model: str = "BAAI/bge-large-en-v1.5"
    embedding_dimensions: int = 1024
    connect_documents: bool = False

    # Database parameters
    graph_database_provider: str = os.getenv("GRAPH_DB_PROVIDER", "NETWORKX")

    if (
        os.getenv("ENV") == "prod"
        or os.getenv("ENV") == "dev"
        or os.getenv("AWS_ENV") == "dev"
        or os.getenv("AWS_ENV") == "prd"
    ):
        load_dotenv()
        logging.info("graph_db_url: %s", os.getenv("GRAPH_DB_URL_PROD"))
        graph_database_url: str = os.getenv("GRAPH_DB_URL_PROD")
        graph_database_username: str = os.getenv("GRAPH_DB_USER")
        graph_database_password: str = os.getenv("GRAPH_DB_PW")
    else:
        logging.info("graph_db_url: %s", os.getenv("GRAPH_DB_URL"))
        graph_database_url: str = os.getenv("GRAPH_DB_URL")
        graph_database_username: str = os.getenv("GRAPH_DB_USER")
        graph_database_password: str = os.getenv("GRAPH_DB_PW")

    weaviate_url: str = os.getenv("WEAVIATE_URL")
    weaviate_api_key: str = os.getenv("WEAVIATE_API_KEY")

    if (
        os.getenv("ENV") == "prod"
        or os.getenv("ENV") == "dev"
        or os.getenv("AWS_ENV") == "dev"
        or os.getenv("AWS_ENV") == "prd"
    ):
        load_dotenv()

        db_host: str = os.getenv("POSTGRES_HOST")
        logging.info("db_host: %s", db_host)
        db_user: str = os.getenv("POSTGRES_USER")
        db_password: str = os.getenv("POSTGRES_PASSWORD")
        db_name: str = os.getenv("POSTGRES_DB")

    # Model parameters and configuration for interlayer scoring
    intra_layer_score_treshold: float = 0.98


    # Client ID
    anon_clientid: Optional[str] = field(default_factory=lambda: uuid.uuid4().hex)

    #Chunking parameters
    chunk_size: int = 1500
    chunk_overlap: int = 0
    chunk_strategy: str = ChunkStrategy.PARAGRAPH

    def load(self):
        """Loads the configuration from a file or environment variables."""
        config = configparser.ConfigParser()
        config.read(self.config_path)

        # Override with environment variables if they exist
        for attr in self.__annotations__:
            env_value = os.getenv(attr.upper())
            if env_value is not None:
                setattr(self, attr, env_value)

        # Load from config file
        if config.sections():
            for section in config.sections():
                for key, value in config.items(section):
                    if hasattr(self, key):
                        setattr(self, key, value)

    def save(self):
        """Saves the current configuration to a file."""
        config = configparser.ConfigParser()

        # Save the current settings to the config file
        for attr, value in self.__dict__.items():
            section, option = attr.split("_", 1)
            if not config.has_section(section):
                config.add_section(section)
            config.set(section, option, str(value))

        with open(self.config_path, "w") as configfile:
            config.write(configfile)

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dictionary representation of the configuration."""
        return {attr: getattr(self, attr) for attr in self.__annotations__}

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> "Config":
        """Creates a Config instance from a dictionary."""
        config = cls()
        for attr, value in config_dict.items():
            if hasattr(config, attr):
                setattr(config, attr, value)
        return config


from uuid import UUID

def encode_uuid(uuid: UUID) -> str:
    uuid_int = uuid.int
    base = 52
    charset = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

    encoded = ''
    while len(encoded) < 36:
        uuid_int, remainder = divmod(uuid_int, base)
        uuid_int = uuid_int * 8
        encoded = charset[remainder] + encoded

    return encoded



from cognee.infrastructure.databases.relational.get_database import get_database

async def is_existing_memory(memory_name: str):
    memory = await (get_database().get_memory_by_name(memory_name))

    return memory is not None



from typing import Protocol, BinaryIO

class IngestionData(Protocol):
    data: str | BinaryIO = None

    def get_data(self):
        raise NotImplementedError()

    def get_identifier(self):
        raise NotImplementedError()

    def get_metadata(self):
        raise NotImplementedError()



from jinja2 import Environment, FileSystemLoader, select_autoescape
from cognee.root_dir import get_absolute_path

def render_prompt(filename: str,  context: dict) -> str:
    """Render a Jinja2 template asynchronously.
    :param filename: The name of the template file to render.
    :param context: The context to render the template with.
    :return: The rendered template as a string."""

    # Set the base directory relative to the cognee root directory
    base_directory = get_absolute_path("./infrastructure/llm/prompts")

    # Initialize the Jinja2 environment to load templates from the filesystem
    env = Environment(
        loader = FileSystemLoader(base_directory),
        autoescape = select_autoescape(["html", "xml", "txt"])
    )

    # Load the template by name
    template = env.get_template(filename)

    # Render the template with the provided context
    rendered_template = template.render(context)

    return rendered_template



from cognee.infrastructure.files.storage import LocalStorage
from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client

class prune():
    @staticmethod
    async def prune_data():
        data_root_directory = infrastructure_config.get_config()["data_root_directory"]
        LocalStorage.remove_all(data_root_directory)

    @staticmethod
    async def prune_system(graph = True, vector = True):
        infra_config = infrastructure_config.get_config()

        if graph:
            graph_client = await get_graph_client(infra_config["graph_engine"])
            await graph_client.delete_graph()

        if vector:
            vector_client = infra_config["vector_engine"]
            await vector_client.prune()


if __name__ == "__main__":
    import asyncio
    async def main():
        await prune.prune_data()
        await prune.prune_system()


    asyncio.run(main())



""" LLM Interface """

from typing import Type, Protocol
from abc import abstractmethod
from pydantic import BaseModel
class LLMInterface(Protocol):
    """ LLM Interface """

    # @abstractmethod
    # async def async_get_embedding_with_backoff(self, text, model="text-embedding-ada-002"):
    #     """To get text embeddings, import/call this function"""
    #     raise NotImplementedError
    #
    # @abstractmethod
    # def get_embedding_with_backoff(self, text: str, model: str = "text-embedding-ada-002"):
    #     """To get text embeddings, import/call this function"""
    #     raise NotImplementedError
    #
    # @abstractmethod
    # async def async_get_batch_embeddings_with_backoff(self, texts: List[str], models: List[str]):
    #     """To get multiple text embeddings in parallel, import/call this function"""
    #     raise NotImplementedError

    # """ Get completions """
    # async def acompletions_with_backoff(self, **kwargs):
    #     raise NotImplementedError
    #
    """ Structured output """
    @abstractmethod
    async def acreate_structured_output(self,
                                        text_input: str,
                                        system_prompt: str,
                                        response_model: Type[BaseModel]) -> BaseModel:
        """To get structured output, import/call this function"""
        raise NotImplementedError

    @abstractmethod
    def show_prompt(self, text_input: str, system_prompt: str) -> str:
        """To get structured output, import/call this function"""
        raise NotImplementedError



""" Fetches the context of a given node in the graph"""
from typing import Union, Dict

from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
import networkx as nx
from cognee.shared.data_models import GraphDBType

async def search_neighbour(graph: Union[nx.Graph, any], id: str, infrastructure_config: Dict,
                           other_param: dict = None):
    """
    Search for nodes that share the same 'layer_uuid' as the specified node and return their descriptions.
    Adapts to both NetworkX graphs and Neo4j graph databases based on the configuration.

    Parameters:
    - graph (Union[nx.Graph, AsyncSession]): The graph object or Neo4j session.
    - id (str): The identifier of the node to match against.
    - infrastructure_config (Dict): Configuration that includes the graph engine type.
    - other_param (dict, optional): A dictionary that may contain 'node_id' to specify the node.

    Returns:
    - List[str]: A list of 'description' attributes of nodes that share the same 'layer_uuid' with the specified node.
    """
    node_id = other_param.get('node_id') if other_param else None

    if node_id is None:
        return []

    if infrastructure_config.get_config()["graph_engine"] == GraphDBType.NETWORKX:
        if isinstance(graph, nx.Graph):
            if node_id not in graph:
                return []

            relevant_context = []
            target_layer_uuid = graph.nodes[node_id].get('layer_uuid')

            for n, attr in graph.nodes(data=True):
                if attr.get('layer_uuid') == target_layer_uuid and 'description' in attr:
                    relevant_context.append(attr['description'])

            return relevant_context
        else:
            raise ValueError("Graph object does not match the specified graph engine type in the configuration.")

    elif infrastructure_config.get_config()["graph_engine"] == GraphDBType.NEO4J:
        if isinstance(graph, AsyncSession):
            cypher_query = """
            MATCH (target {id: $node_id})
            WITH target.layer_uuid AS layer
            MATCH (n)
            WHERE n.layer_uuid = layer AND EXISTS(n.description)
            RETURN n.description AS description
            """
            result = await graph.run(cypher_query, node_id=node_id)
            descriptions = [record["description"] for record in await result.list()]

            return descriptions
        else:
            raise ValueError("Graph session does not match the specified graph engine type in the configuration.")

    else:
        raise ValueError("Unsupported graph engine type in the configuration.")



# if __name__ == '__main__':
#     import asyncio
#     async def main():
#         from cognee.shared.data_models import GraphDBType
#
#         graph_client = get_graph_client(GraphDBType.NETWORKX)
#         graph = await  graph_client.graph
#
#         await fetch_context(graph, "1")
#
#     asyncio.run(main())








from typing import Union, Dict
import networkx as nx
from cognee.shared.data_models import GraphDBType

async def search_summary(graph: Union[nx.Graph, any], query: str, infrastructure_config: Dict, other_param: str = None) -> Dict[str, str]:
    """
    Filter nodes based on a condition (such as containing 'SUMMARY' in their identifiers) and return their summary attributes.
    Supports both NetworkX graphs and Neo4j graph databases based on the configuration.

    Parameters:
    - graph (Union[nx.Graph, AsyncSession]): The graph object or Neo4j session.
    - query (str): The query string to filter nodes by, e.g., 'SUMMARY'.
    - infrastructure_config (Dict): Configuration that includes the graph engine type.
    - other_param (str, optional): An additional parameter, unused in this implementation but could be for future enhancements.

    Returns:
    - Dict[str, str]: A dictionary where keys are node identifiers containing the query string, and values are their 'summary' attributes.
    """
    if infrastructure_config.get_config()["graph_engine"] == GraphDBType.NETWORKX:
        return {node: data.get('summary') for node, data in graph.nodes(data=True) if query in node and 'summary' in data}

    elif infrastructure_config.get_config()["graph_engine"] == GraphDBType.NEO4J:
        cypher_query = f"""
        MATCH (n)
        WHERE n.id CONTAINS $query AND EXISTS(n.summary)
        RETURN n.id AS nodeId, n.summary AS summary
        """
        results = await graph.run(cypher_query, query=query)
        summary_data = {record["nodeId"]: record["summary"] for record in await results.list()}
        return summary_data

    else:
        raise ValueError("Unsupported graph engine type in the configuration.")



""" Neo4j Adapter for Graph Database"""
import json
import logging
from typing import Optional, Any, List, Dict
from contextlib import asynccontextmanager
from neo4j import AsyncSession
from neo4j import AsyncGraphDatabase
from neo4j.exceptions import Neo4jError
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface

logger = logging.getLogger("Neo4jAdapter")

class Neo4jAdapter(GraphDBInterface):
    def __init__(
        self,
        graph_database_url: str,
        graph_database_username: str,
        graph_database_password: str,
        driver: Optional[Any] = None,
    ):
        self.driver = driver or AsyncGraphDatabase.driver(
            graph_database_url,
            auth = (graph_database_username, graph_database_password)
        )

    async def close(self) -> None:
        await self.driver.close()

    @asynccontextmanager
    async def get_session(self) -> AsyncSession:
        async with self.driver.session() as session:
            yield session

    async def query(
        self,
        query: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, Any]]:
        try:
            async with self.get_session() as session:
                result = await session.run(query, parameters=params)
                data = await result.data()
                await self.close()
                return data
        except Neo4jError as error:
            logger.error("Neo4j query error: %s", error, exc_info = True)
            raise error

    async def graph(self):
        return await self.get_session()

    async def add_node(self, node_id: str, node_properties: Dict[str, Any] = None):
        node_id = node_id.replace(":", "_")

        serialized_properties = self.serialize_properties(node_properties)

        if "name" not in serialized_properties:
            serialized_properties["name"] = node_id

        # serialized_properties["created_at"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        # serialized_properties["updated_at"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # properties = ", ".join(f"{property_name}: ${property_name}" for property_name in serialized_properties.keys())

        query = f"""MERGE (node:`{node_id}` {{id: $node_id}})
                ON CREATE SET node += $properties
                RETURN ID(node) AS internal_id, node.id AS nodeId"""

        params = {
            "node_id": node_id,
            "properties": serialized_properties,
        }

        return await self.query(query, params)

    async def add_nodes(self, nodes: list[tuple[str, dict[str, Any]]]) -> None:
        # nodes_data = []

        for node in nodes:
            node_id, node_properties = node
            node_id = node_id.replace(":", "_")

            await self.add_node(
                node_id = node_id,
                node_properties = node_properties,
            )


        #     serialized_properties = self.serialize_properties(node_properties)

        #     if "name" not in serialized_properties:
        #         serialized_properties["name"] = node_id

        #     nodes_data.append({
        #         "node_id": node_id,
        #         "properties": serialized_properties,
        #     })

        # query = """UNWIND $nodes_data AS node_data
        #         MERGE (node:{id: node_data.node_id})
        #         ON CREATE SET node += node_data.properties
        #         RETURN ID(node) AS internal_id, node.id AS id"""

        # params = {"nodes_data": nodes_data}

        # result = await self.query(query, params)

        # await self.close()

        # return result

    async def extract_node_description(self, node_id: str):
        query = """MATCH (n)-[r]->(m)
                    WHERE n.id = $node_id
                    AND NOT m.id CONTAINS 'DefaultGraphModel'
                    RETURN m
                    """

        result = await self.query(query, dict(node_id = node_id))

        descriptions = []

        for node in result:
            # Assuming 'm' is a consistent key in your data structure
            attributes = node.get("m", {})

            # Ensure all required attributes are present
            if all(key in attributes for key in ["id", "layer_id", "description"]):
                descriptions.append({
                    "id": attributes["id"],
                    "layer_id": attributes["layer_id"],
                    "description": attributes["description"],
                })

        return descriptions

    async def get_layer_nodes(self):
        query = """MATCH (node) WHERE node.layer_id IS NOT NULL
        RETURN node"""

        return [result['node'] for result in (await self.query(query))]

    async def extract_node(self, node_id: str):
        query= """
        MATCH(node {id: $node_id})
        RETURN node
        """

        results = [node['node'] for node in (await self.query(query, dict(node_id = node_id)))]

        return results[0] if len(results) > 0 else None

    async def delete_node(self, node_id: str):
        node_id = id.replace(":", "_")

        query = f"MATCH (node:`{node_id}` {{id: $node_id}}) DETACH DELETE n"
        params = { "node_id": node_id }

        return await self.query(query, params)

    async def add_edge(self, from_node: str, to_node: str, relationship_name: str, edge_properties: Optional[Dict[str, Any]] = {}):
        serialized_properties = self.serialize_properties(edge_properties)
        from_node = from_node.replace(":", "_")
        to_node = to_node.replace(":", "_")

        query = f"""MATCH (from_node:`{from_node}` {{id: $from_node}}), (to_node:`{to_node}` {{id: $to_node}})
                MERGE (from_node)-[r:`{relationship_name}`]->(to_node)
                SET r += $properties
                RETURN r"""

        params = {
            "from_node": from_node,
            "to_node": to_node,
            "properties": serialized_properties
        }

        return await self.query(query, params)


    async def add_edges(self, edges: list[tuple[str, str, str, dict[str, Any]]]) -> None:
        # edges_data = []

        for edge in edges:
            from_node, to_node, relationship_name, edge_properties = edge
            from_node = from_node.replace(":", "_")
            to_node = to_node.replace(":", "_")

            await self.add_edge(
                from_node = from_node,
                to_node = to_node,
                relationship_name = relationship_name,
                edge_properties = edge_properties
            )

            # Filter out None values and do not serialize; Neo4j can handle complex types like arrays directly
        #     serialized_properties = self.serialize_properties(edge_properties)

        #     edges_data.append({
        #         "from_node": from_node,
        #         "to_node": to_node,
        #         "relationship_name": relationship_name,
        #         "properties": serialized_properties
        #     })

        # query = """UNWIND $edges_data AS edge_data
        #         MATCH (from_node:{id: edge_data.from_node}), (to_node:{id: edge_data.to_node})
        #         MERGE (from_node)-[r:{edge_data.relationship_name}]->(to_node)
        #         ON CREATE SET r += edge_data.properties
        #         RETURN r"""

        # params = {"edges_data": edges_data}

        # result = await self.query(query, params)

        # await self.close()

        # return result


    async def filter_nodes(self, search_criteria):
        query = f"""MATCH (node)
                WHERE node.id CONTAINS '{search_criteria}'
                RETURN node"""


        return await self.query(query)


    async def delete_graph(self):
        query = """MATCH (node)
                DETACH DELETE node;"""

        return await self.query(query)

    def serialize_properties(self, properties = dict()):
        return {
            property_key: json.dumps(property_value)
            if isinstance(property_value, (dict, list))
            else property_value for property_key, property_value in properties.items()
        }



import os
import shutil
from typing import BinaryIO, Union
from .StorageManager import Storage

class LocalStorage(Storage):
    storage_path: str = None

    def __init__(self, storage_path: str):
        self.storage_path = storage_path

    def store(self, file_path: str, data: Union[BinaryIO, str]):
        full_file_path = self.storage_path + "/" + file_path

        LocalStorage.ensure_directory_exists(self.storage_path)

        with open(
            full_file_path,
            mode = "w" if isinstance(data, str) else "wb",
            encoding = "utf-8" if isinstance(data, str) else None
        ) as f:
            f.write(data if isinstance(data, str) else data.read())

    def retrieve(self, file_path: str, mode: str = "rb"):
        full_file_path = self.storage_path + "/" + file_path

        with open(full_file_path, mode = mode) as f:
            return f.read()

    @staticmethod
    def ensure_directory_exists(file_path: str):
        if not os.path.exists(file_path):
            os.makedirs(file_path)

    def remove(self, file_path: str):
        os.remove(self.storage_path + "/" + file_path)

    @staticmethod
    def copy_file(source_file_path: str, destination_file_path: str):
        return shutil.copy2(source_file_path, destination_file_path)

    @staticmethod
    def remove_all(tree_path: str):
        try:
            shutil.rmtree(tree_path)
        except FileNotFoundError:
            pass



from uuid import uuid4
from typing import List
from datetime import datetime
from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.databases.vector import DataPoint

async def add_label_nodes(graph_client, parent_node_id: str, chunk_id: str, keywords: List[str]) -> None:
    vector_client = infrastructure_config.get_config("vector_engine")

    keyword_nodes = []

    for keyword in keywords:
        keyword_id = f"DATA_LABEL_{keyword.upper().replace(' ', '_')}"

        keyword_nodes.append((
            keyword_id,
            dict(
                id = keyword_id,
                chunk_id = chunk_id,
                name = keyword.lower().capitalize(),
                keyword = keyword.lower(),
                entity_type = "Keyword",
                created_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                updated_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            ),
        ))

    # Add data to graph
    await graph_client.add_nodes(keyword_nodes)
    await graph_client.add_edges([
        (
            parent_node_id,
            keyword_id,
            "refers_to",
            dict(relationship_name = "refers_to"),
        ) for (keyword_id, __) in keyword_nodes
    ])

    # Add data to vector
    keyword_data_points = [
        DataPoint(
            id = str(uuid4()),
            payload = dict(
                value = keyword_data["keyword"],
                references = dict(
                    node_id = keyword_node_id,
                    cognitive_layer = parent_node_id,
                ),
            ),
            embed_field = "value"
        ) for (keyword_node_id, keyword_data) in keyword_nodes
    ]

    try:
        await vector_client.create_collection(parent_node_id)
    except Exception:
        # It's ok if the collection already exists.
        pass

    await vector_client.create_data_points(parent_node_id, keyword_data_points)



from cognee.shared.data_models import Document
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface

async def add_document_node(graph_client: GraphDBInterface, parent_node_id, document_metadata):
    document_id = f"DOCUMENT__{document_metadata['id']}"

    document = await graph_client.extract_node(document_id)

    if not document:
        document = Document(
            id = document_id,
            title = document_metadata["name"],
            file_path = document_metadata["file_path"],
        ).model_dump()

    document["entity_type"] = "Document"

    await graph_client.add_node(document_id, document)

    await graph_client.add_edge(parent_node_id, document_id, "has_document", dict(relationship_name = "has_document"))

    return document_id



import dsp
import dspy
from dspy.evaluate.evaluate import Evaluate
from dspy.primitives.example import Example
from cognee.modules.data.extraction.knowledge_graph.extract_knowledge_graph_module import ExtractKnowledgeGraph
from cognee.root_dir import get_absolute_path
from cognee.config import Config
from cognee.shared.data_models import Answer
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.modules.cognify.dataset import HotPotQA

config = Config()
config.load()

def evaluate():
    dataset = HotPotQA(
        train_seed = 1,
        train_size = 16,
        eval_seed = 2023,
        dev_size = 8,
        test_size = 0,
        keep_details = True,
    )

    #Evaluate
    evaluate_examples = [
        Example(
            base = None,
            question = None,
            context = "\r\n".join("".join(sentences) for sentences in example.context["sentences"]),
            answer = example.answer,
        ) for example in dataset.dev
    ]

    devset = [example.with_inputs("context", "question") for example in evaluate_examples]

    evaluate_on_hotpotqa = Evaluate(devset = devset, num_threads = 1, display_progress = True, display_table = 5, max_tokens = 4096)

    gpt4 = dspy.OpenAI(model = config.openai_model, api_key = config.openai_key, model_type = "chat", max_tokens = 4096)
    compiled_extract_knowledge_graph = ExtractKnowledgeGraph(lm = gpt4)
    compiled_extract_knowledge_graph.load(get_absolute_path("./programs/extract_knowledge_graph/extract_knowledge_graph.json"))

    def evaluate_answer(example, graph_prediction, trace = None):
        llm_client = get_llm_client()

        try:
            answer_prediction = llm_client.create_structured_output(
                text_input = example.question,
                system_prompt = f"""Answer the question by looking at the provided knowledge graph.
                Use only the graph to answer the question and be very brief.
                This is the knowledge graph:
                {graph_prediction.graph.model_dump(mode = "json")}""",
                response_model = Answer,
            )
        except:
            return False

        return dsp.answer_match(example.answer, [answer_prediction.answer], frac = 0.8) or \
            dsp.passage_match([example.answer], [answer_prediction.answer])

    gpt4 = dspy.OpenAI(model = config.openai_model, api_key = config.openai_key, model_type = "chat", max_tokens = 4096)
    dspy.settings.configure(lm = gpt4)

    evaluate_on_hotpotqa(compiled_extract_knowledge_graph, metric = evaluate_answer)


if __name__ == "__main__":
    evaluate()



from typing import List

async def add_classification_nodes(graph_client, parent_node_id: str, categories: List) -> None:
    for category in categories:
        data_type = category["data_type"].upper().replace(' ', '_')
        category_name = category["category_name"].upper().replace(' ', '_').replace("'", "").replace("/", "_")

        data_type_node_id = f"DATA_TYPE__{data_type}"

        data_type_node = await graph_client.extract_node(data_type_node_id)

        if not data_type_node:
            data_type_node = await graph_client.add_node(data_type_node_id, dict(name = data_type, entity_type = "DataType"))

        await graph_client.add_edge(data_type_node_id, parent_node_id, "classified_as", dict(relationship_name = "classified_as"))

        category_node_id = f"DATA_CATEGORY__{category_name}"

        category_node = await graph_client.extract_node(category_node_id)

        if not category_node:
            category_node = await graph_client.add_node(category_node_id, dict(name = category_name, entity_type = "DataCategory"))

        await graph_client.add_edge(category_node_id, parent_node_id, "classified_as", dict(relationship_name = "classified_as"))



from typing import List

async def add_cognitive_layers(graph_client, parent_node_id: str, cognitive_layers: List):
    cognitive_layer_nodes = [(
        generate_cognitive_layer_id(cognitive_layer.name),
        dict(
            id = generate_cognitive_layer_id(cognitive_layer.name),
            name = cognitive_layer.name,
            description = cognitive_layer.description,
        ),
    ) for cognitive_layer in cognitive_layers]

    await graph_client.add_nodes(cognitive_layer_nodes)
    await graph_client.add_edges((
        parent_node_id,
        cognitive_layer_id,
        "decomposed_as",
        dict(relationship_name = "decomposed_as"),
    ) for (cognitive_layer_id, _) in cognitive_layer_nodes)

    return cognitive_layer_nodes

def generate_cognitive_layer_id(layer_id: str) -> str:
    return f"COGNITIVE_LAYER__{layer_id.upper().replace(' ', '_')}".replace("'", "").replace("/", "_")



""" This module is used to set the configuration of the system."""
from cognee.infrastructure import infrastructure_config

class config():
    @staticmethod
    def system_root_directory(system_root_directory: str):
        infrastructure_config.set_config({
            "system_root_directory": system_root_directory
        })

    @staticmethod
    def data_root_directory(data_root_directory: str):
        infrastructure_config.set_config({
            "data_root_directory": data_root_directory
        })

    @staticmethod
    def set_classification_model(classification_model: object):
        infrastructure_config.set_config({
            "classification_model": classification_model
        })

    @staticmethod
    def set_summarization_model(summarization_model: object):
        infrastructure_config.set_config({
            "summarization_model": summarization_model
        })

    @staticmethod
    def set_labeling_model(labeling_model: object):
        infrastructure_config.set_config({
            "labeling_model": labeling_model
        })

    @staticmethod
    def set_graph_model(graph_model: object):
        infrastructure_config.set_config({
            "graph_model": graph_model
        })

    @staticmethod
    def set_cognitive_layer_model(cognitive_layer_model: object):
        infrastructure_config.set_config({
            "cognitive_layer_model": cognitive_layer_model
        })

    @staticmethod
    def set_graph_engine(graph_engine: object):
        infrastructure_config.set_config({
            "graph_engine": graph_engine
        })

    @staticmethod
    def llm_provider(llm_provider: str):
        infrastructure_config.set_config({
            "llm_provider": llm_provider
        })

    @staticmethod
    def intra_layer_score_treshold(intra_layer_score_treshold: str):
        infrastructure_config.set_config({
            "intra_layer_score_treshold": intra_layer_score_treshold
        })

    @staticmethod
    def connect_documents(connect_documents: bool):
        infrastructure_config.set_config({
            "connect_documents": connect_documents
        })
    @staticmethod
    def set_chunk_strategy(chunk_strategy: object):
        infrastructure_config.set_config({
            "chunk_strategy": chunk_strategy
        })



import logging
import asyncio
from cognee.infrastructure import infrastructure_config
from .extraction.knowledge_graph.extract_knowledge_graph import extract_knowledge_graph

logger = logging.getLogger(__name__)

async def get_layer_graphs(content: str, cognitive_layers: list[tuple[str, dict]]):
    try:
        graph_awaitables = [
            extract_knowledge_graph(
                content,
                cognitive_layer_data["name"],
                infrastructure_config.get_config()["graph_model"]
            ) for (_, cognitive_layer_data) in cognitive_layers
        ]

        graphs = await asyncio.gather(*graph_awaitables)

        return [(layer_id, graphs[layer_index]) for (layer_index, (layer_id, __)) in enumerate(cognitive_layers)]
    except Exception as error:
        logger.error("Error extracting graph from content: %s", error, exc_info = True)
        raise error



from typing import Protocol, BinaryIO

class Storage(Protocol):
    def store(self, file_path: str, data: bytes):
        pass

    def retrieve(self, file_path: str):
        pass

    def remove(self, file_path: str):
        pass

class StorageManager():
    storage: Storage = None

    def __init__ (self, storage: Storage):
        self.storage = storage

    def store(self, file_path: str, data: BinaryIO):
        return self.storage.store(file_path, data)

    def retrieve(self, file_path: str):
        return self.storage.retrieve(file_path)

    def remove(self, file_path: str):
        return self.storage.remove(file_path)



from .api.v1.config.config import config
from .api.v1.add.add import add
from .api.v1.cognify.cognify import cognify
from .api.v1.datasets.datasets import datasets
from .api.v1.search.search import search, SearchType
from .api.v1.prune import prune



from typing import List
from enum import Enum
from cognee.modules.users.memory import create_information_points, is_existing_memory

class MemoryType(Enum):
    GRAPH = "GRAPH"
    VECTOR = "VECTOR"
    RELATIONAL = "RELATIONAL"

class MemoryException(Exception):
    message: str

    def __init__(self, message: str):
        self.message = message


async def remember(user_id: str, memory_name: str, payload: List[str]):
    if await is_existing_memory(memory_name) is False:
        raise MemoryException(f"Memory with the name \"{memory_name}\" doesn't exist.")

    await create_information_points(memory_name, payload)


""" FastAPI server for the Cognee API. """
from uuid import UUID

import uvicorn
import logging

# Set up logging
logging.basicConfig(
    level=logging.INFO,  # Set the logging level (e.g., DEBUG, INFO, WARNING, ERROR, CRITICAL)
    format="%(asctime)s [%(levelname)s] %(message)s",  # Set the log message format
)

logger = logging.getLogger(__name__)

from cognee.config import Config

config = Config()
config.load()

from typing import Dict, Any, List, Union, BinaryIO
from fastapi import FastAPI
from fastapi.responses import JSONResponse
from pydantic import BaseModel

app = FastAPI(debug=True)
#
# from auth.cognito.JWTBearer import JWTBearer
# from auth.auth import jwks
#
# auth = JWTBearer(jwks)


@app.get("/")
async def root():
    """
    Root endpoint that returns a welcome message.
    """
    return {"message": "Hello, World, I am alive!"}


@app.get("/health")
def health_check():
    """
    Health check endpoint that returns the server status.
    """
    return {"status": "OK"}


class Payload(BaseModel):
    payload: Dict[str, Any]


class AddPayload(BaseModel):
    data: Union[str, BinaryIO, List[Union[str, BinaryIO]]]
    dataset_id: UUID
    dataset_name: str
    class Config:
        arbitrary_types_allowed = True # This is required to allow the use of Union
class CognifyPayload(BaseModel):
    datasets: Union[str, List[str]]

class SearchPayload(BaseModel):
    query_params:  Dict[str, Any]
@app.post("/add", response_model=dict)
async def add(payload: AddPayload):
    """ This endpoint is responsible for adding data to the graph."""
    from v1.add.add import add

    try:
        await add(
            payload.data,
            payload.dataset_name,
        )
    except Exception as error:
        return JSONResponse(
            status_code = 409,
            content = { "error": error }
        )

@app.post("/cognify", response_model=dict)
async def cognify(payload: CognifyPayload):
    """ This endpoint is responsible for the cognitive processing of the content."""
    from v1.cognify.cognify import cognify

    try:
        await cognify(payload.datasets)
    except Exception as error:
        return JSONResponse(
            status_code = 409,
            content = { "error": error }
        )


@app.post("/search", response_model=dict)
async def search(payload: SearchPayload):
    """ This endpoint is responsible for searching for nodes in the graph."""
    from v1.search.search import search

    try:
        search_type = 'SIMILARITY'
        await search(search_type, payload.query_params)
    except Exception as error:
        return JSONResponse(
            status_code = 409,
            content = { "error": error }
        )



def start_api_server(host: str = "0.0.0.0", port: int = 8000):
    """
    Start the API server using uvicorn.
    Parameters:
    host (str): The host for the server.
    port (int): The port for the server.
    """
    try:
        logger.info(f"Starting server at {host}:{port}")
        uvicorn.run(app, host=host, port=port)
    except Exception as e:
        logger.exception(f"Failed to start server: {e}")
        # Here you could add any cleanup code or error recovery code.


if __name__ == "__main__":
    start_api_server()



from datetime import datetime
from uuid import uuid4
from typing import List, Tuple, TypedDict
from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.databases.vector import DataPoint
from cognee.shared.data_models import KnowledgeGraph
from cognee.utils import extract_pos_tags, extract_named_entities, extract_sentiment_vader

class GraphLike(TypedDict):
    nodes: List
    edges: List


async def add_cognitive_layer_graphs(
    graph_client,
    chunk_collection: str,
    chunk_id: str,
    layer_graphs: List[Tuple[str, GraphLike]],
):
    vector_client = infrastructure_config.get_config("vector_engine")

    for (layer_id, layer_graph) in layer_graphs:
        graph_nodes = []
        graph_edges = []

        if not isinstance(layer_graph, KnowledgeGraph):
            layer_graph = KnowledgeGraph.parse_obj(layer_graph)

        for node in layer_graph.nodes:
            node_id = generate_proposition_node_id(node.id)

            entity_type_node_id = generate_type_node_id(node.entity_type)
            entity_type_node = await graph_client.extract_node(entity_type_node_id)

            if not entity_type_node:
                node_name = node.entity_type.lower().capitalize()
              
                entity_type_node = (
                    entity_type_node_id,
                    dict(
                        id = entity_type_node_id,
                        name = node_name,
                        entity_type = node_name,
                        created_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        updated_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    )
                )

                graph_nodes.append(entity_type_node)

                # Add relationship between document and entity type: "Document contains Person"
                graph_edges.append((
                    layer_id,
                    entity_type_node_id,
                    "contains",
                    dict(relationship_name = "contains"),
                ))

            pos_tags = extract_pos_tags(node.entity_description)
            named_entities = extract_named_entities(node.entity_description)
            sentiment = extract_sentiment_vader(node.entity_description)

            graph_nodes.append((
                node_id,
                dict(
                    id = node_id,
                    layer_id = layer_id,
                    chunk_id = chunk_id,
                    chunk_collection = chunk_collection,
                    name = node.entity_name,
                    entity_type = node.entity_type.lower().capitalize(),
                    description = node.entity_description,
                    pos_tags = pos_tags,
                    sentiment = sentiment,
                    named_entities = named_entities,
                    created_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    updated_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                )
            ))

            # Add relationship between entity type and entity itself: "Jake is Person"
            graph_edges.append((
                node_id,
                entity_type_node_id,
                "is",
                dict(relationship_name = "is"),
            ))

            graph_edges.append((
                layer_id,
                node_id,
                "contains",
                dict(relationship_name = "contains"),
            ))

        # Add relationship that came from graphs.
        for edge in layer_graph.edges:
            graph_edges.append((
                generate_proposition_node_id(edge.source_node_id),
                generate_proposition_node_id(edge.target_node_id),
                edge.relationship_name,
                dict(relationship_name = edge.relationship_name),
            ))

        await graph_client.add_nodes(graph_nodes)

        await graph_client.add_edges(graph_edges)

        try:
            await vector_client.create_collection(layer_id)
        except Exception:
            # It's ok if the collection already exists.
            pass

        data_points = [
            DataPoint(
                id = str(uuid4()),
                payload = dict(
                    value = node_data["name"],
                    references = dict(
                        node_id = node_id,
                        cognitive_layer = layer_id,
                    ),
                ),
                embed_field = "value"
            ) for (node_id, node_data) in graph_nodes
        ]

        await vector_client.create_data_points(layer_id, data_points)


def generate_proposition_node_id(node_id: str) -> str:
    return f"PROPOSITION_NODE__{node_id.upper().replace(' ', '_')}".replace("'", "")

def generate_type_node_id(node_id: str) -> str:
    return f"PROPOSITION_TYPE_NODE__{node_id.upper().replace(' ', '_')}".replace("'", "")


import logging
from cognee.infrastructure import infrastructure_config
from .extraction.extract_categories import extract_categories

logger = logging.getLogger(__name__)

async def get_content_categories(content: str):
    try:
        return await extract_categories(
            content,
            infrastructure_config.get_config()["classification_model"]
        )
    except Exception as error:
        logger.error("Error extracting categories from content: %s", error, exc_info = True)
        raise error



from uuid import uuid5, UUID
from .data_types import IngestionData

null_uuid: UUID = UUID("00000000-0000-0000-0000-000000000000")

def identify(data: IngestionData) -> str:
    data_id: str = data.get_identifier()

    return str(uuid5(null_uuid, data_id)).replace("-", "")



import asyncio
from uuid import uuid4
from typing import List, Union
import logging
import instructor
from openai import OpenAI
from nltk.corpus import stopwords
from cognee.config import Config
from cognee.modules.cognify.graph.add_data_chunks import add_data_chunks
from cognee.modules.cognify.graph.add_document_node import add_document_node
from cognee.modules.cognify.graph.add_classification_nodes import add_classification_nodes
from cognee.modules.cognify.graph.add_cognitive_layer_graphs import add_cognitive_layer_graphs
from cognee.modules.cognify.graph.add_summary_nodes import add_summary_nodes
from cognee.modules.cognify.graph.add_node_connections import group_nodes_by_layer, \
    graph_ready_output, connect_nodes_in_graph
from cognee.modules.cognify.llm.resolve_cross_graph_references import resolve_cross_graph_references
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
from cognee.modules.cognify.graph.add_label_nodes import add_label_nodes
from cognee.modules.cognify.graph.add_cognitive_layers import add_cognitive_layers
# from cognee.modules.cognify.graph.initialize_graph import initialize_graph
from cognee.infrastructure.files.utils.guess_file_type import guess_file_type, FileTypeException
from cognee.infrastructure.files.utils.extract_text_from_file import extract_text_from_file
from cognee.infrastructure import infrastructure_config
from cognee.modules.data.get_content_categories import get_content_categories
from cognee.modules.data.get_content_summary import get_content_summary
from cognee.modules.data.get_cognitive_layers import get_cognitive_layers
from cognee.modules.data.get_layer_graphs import get_layer_graphs
from cognee.utils import send_telemetry


config = Config()
config.load()

aclient = instructor.patch(OpenAI())

USER_ID = "default_user"

logger = logging.getLogger("cognify")

async def cognify(datasets: Union[str, List[str]] = None):
    """This function is responsible for the cognitive processing of the content."""
    # Has to be loaded in advance, multithreading doesn't work without it.
    stopwords.ensure_loaded()

    db_engine = infrastructure_config.get_config()["database_engine"]

    if datasets is None or len(datasets) == 0:
        datasets = db_engine.get_datasets()

    awaitables = []

    # datasets is a list of dataset names
    if isinstance(datasets, list):
        for dataset in datasets:
            awaitables.append(cognify(dataset))

        graphs = await asyncio.gather(*awaitables)
        return graphs[0]

    added_datasets = db_engine.get_datasets()

    dataset_files = []
    # datasets is a dataset name string
    dataset_name = datasets.replace(".", "_").replace(" ", "_")

    for added_dataset in added_datasets:
        if dataset_name in added_dataset:
            dataset_files.append((added_dataset, db_engine.get_files_metadata(added_dataset)))

    awaitables = []

    graph_db_type = infrastructure_config.get_config()["graph_engine"]

    graph_client = await get_graph_client(graph_db_type)

    # await initialize_graph(USER_ID, graph_data_model, graph_client)

    data_chunks = {}

    chunk_engine = infrastructure_config.get_config()["chunk_engine"]
    chunk_strategy = infrastructure_config.get_config()["chunk_strategy"]

    for (dataset_name, files) in dataset_files:
        for file_metadata in files:
            with open(file_metadata["file_path"], "rb") as file:
                try:
                    file_type = guess_file_type(file)
                    text = extract_text_from_file(file, file_type)
                    subchunks = chunk_engine.chunk_data(chunk_strategy, text, config.chunk_size, config.chunk_overlap)

                    if dataset_name not in data_chunks:
                        data_chunks[dataset_name] = []

                    for subchunk in subchunks:
                        data_chunks[dataset_name].append(dict(text = subchunk, chunk_id = str(uuid4()), file_metadata = file_metadata))
                except FileTypeException:
                    logger.warning("File (%s) has an unknown file type. We are skipping it.", file_metadata["id"])

    added_chunks: list[tuple[str, str, dict]] = await add_data_chunks(data_chunks)

    await asyncio.gather(
        *[process_text(chunk["collection"], chunk["chunk_id"], chunk["text"], chunk["file_metadata"]) for chunk in added_chunks]
    )

    return graph_client.graph

async def process_text(chunk_collection: str, chunk_id: str, input_text: str, file_metadata: dict):
    print(f"Processing chunk ({chunk_id}) from document ({file_metadata['id']}).")

    graph_client = await get_graph_client(infrastructure_config.get_config()["graph_engine"])

    document_id = await add_document_node(
        graph_client,
        parent_node_id = f"DefaultGraphModel__{USER_ID}", #make a param of defaultgraph model to make sure when user passes his stuff, it doesn't break pipeline
        document_metadata = file_metadata,
    )

    await add_label_nodes(graph_client, document_id, chunk_id, file_metadata["keywords"].split("|"))

    classified_categories = await get_content_categories(input_text)
    await add_classification_nodes(
        graph_client,
        parent_node_id = document_id,
        categories = classified_categories,
    )

    print(f"Chunk ({chunk_id}) classified.")

    content_summary = await get_content_summary(input_text)
    await add_summary_nodes(graph_client, document_id, content_summary)

    print(f"Chunk ({chunk_id}) summarized.")

    cognitive_layers = await get_cognitive_layers(input_text, classified_categories)
    cognitive_layers = (await add_cognitive_layers(graph_client, document_id, cognitive_layers))[:2]

    layer_graphs = await get_layer_graphs(input_text, cognitive_layers)
    await add_cognitive_layer_graphs(graph_client, chunk_collection, chunk_id, layer_graphs)

    if infrastructure_config.get_config()["connect_documents"] is True:
        db_engine = infrastructure_config.get_config()["database_engine"]
        relevant_documents_to_connect = db_engine.fetch_cognify_data(excluded_document_id = file_metadata["id"])

        list_of_nodes = []

        relevant_documents_to_connect.append({
            "layer_id": document_id,
        })

        for document in relevant_documents_to_connect:
            node_descriptions_to_match = await graph_client.extract_node_description(document["layer_id"])
            list_of_nodes.extend(node_descriptions_to_match)

        nodes_by_layer = await group_nodes_by_layer(list_of_nodes)

        results = await resolve_cross_graph_references(nodes_by_layer)

        relationships = graph_ready_output(results)

        await connect_nodes_in_graph(
            graph_client,
            relationships,
            score_threshold = infrastructure_config.get_config()["intra_layer_score_treshold"]
        )

    send_telemetry( "COGNEE_COGNIFY")

    print(f"Chunk ({chunk_id}) cognified.")



""" This module is responsible for creating a semantic graph """
from typing import  Optional, Any
from pydantic import BaseModel

from cognee.infrastructure import infrastructure_config
from cognee.shared.data_models import GraphDBType


async def generate_node_id(instance: BaseModel) -> str:
    for field in ["id", "doc_id", "location_id", "type_id", "node_id"]:
        if hasattr(instance, field):
            return f"{instance.__class__.__name__}:{getattr(instance, field)}"
    return f"{instance.__class__.__name__}_default"


async def add_node(client, parent_id: Optional[str], node_id: str, node_data: dict) -> Optional[Any]:
    """
    Asynchronously adds a node to a graph database and, if applicable, an edge from a parent node to this new node.

    Parameters:
    - client: The database client used to interact with the graph database.
    - parent_id (Optional[str]): The unique identifier of the parent node. If provided, an edge will be added from the parent to the new node.
    - node_id (str): The unique identifier of the new node. If this is "Relationship_default", the node will not be added.
    - node_data (dict): A dictionary containing the data to be stored in the new node.

    Returns:
    - The result of the node addition operation if successful, otherwise None. This could be the newly added node object, a confirmation message, or any relevant data returned by the database client.

    Raises:
    - Exception: If there is an error during the node or edge addition process, it logs the error and continues without interrupting the execution flow.

    Note:
    - The function currently supports adding edges only if the graph database engine is NETWORKX, as specified in the global `infrastructure_config`.
    """

    # Initialize result to None to ensure a clear return path
    result = None

    # Proceed only if the node_id is not meant for default relationships
    if node_id != "Relationship_default":
        try:
            # Attempt to add the node to the graph database
            result = await client.add_node(node_id, node_properties = node_data)

            # Add an edge if a parent ID is provided and the graph engine is NETWORKX
            if parent_id and "default_relationship" in node_data and infrastructure_config.get_config()["graph_engine"] == GraphDBType.NETWORKX:
                await client.add_edge(parent_id, node_id, relationship_name = node_data["default_relationship"]["type"], edge_properties = node_data)
        except Exception as e:
            # Log the exception; consider a logging framework for production use
            print(f"Error adding node or edge: {e}")
            # Depending on requirements, you may want to handle the exception differently

    return result



# async def add_edge(client, parent_id: Optional[str], node_id: str, node_data: Dict[str, Any], created_node_ids: List[Dict[str, str]]) -> None:
#     """
#     Asynchronously adds an edge between two nodes in a graph based on provided node data and identifiers.
#
#     This function is specifically designed to handle cases where the edge represents a default relationship,
#     identified by 'node_id' being 'Relationship_default'. It searches for source and target nodes based on
#     'source' and 'target' keys in the provided 'node_data' dictionary, within a list of previously created nodes.
#
#     Parameters:
#     - client: The graph database client instance used to interact with the graph.
#     - parent_id (Optional[str]): The unique identifier of the parent node. This is currently not used in the function, but may be relevant for extended functionality.
#     - node_id (str): The identifier for the node. The function specifically checks for 'Relationship_default' to proceed with edge creation.
#     - node_data (Dict[str, Any]): A dictionary containing data related to the node, including keys for 'source', 'target', and 'type' to construct the edge.
#     - created_node_ids (List[Dict[str, str]]): A list of dictionaries, each containing 'nodeId' keys representing nodes previously added to the graph.
#
#     Returns:
#     None: The function performs an asynchronous operation to add an edge to the graph but does not return a value.
#
#     Raises:
#     - This function currently does not explicitly raise exceptions but relies on the calling context to handle exceptions raised by 'client.add_edge'.
#
#     Example:
#     ```
#     await add_edge(client, None, "Relationship_default", {'source': 'Node1', 'target': 'Node2', 'type': 'CONNECTS'}, [{'nodeId': 'Node1'}, {'nodeId': 'Node2'}])
#     ```
#     """
#     if node_id == "Relationship_default" and parent_id:
#         # Extract source and target node keys in lower case for case-insensitive matching
#         source_key = node_data.get('source', '').lower()
#         target_key = node_data.get('target', '').lower()
#
#         # Find the first matching node IDs for source and target using case-insensitive comparison
#         source = next((node['nodeId'] for node in created_node_ids if source_key in node['nodeId'].lower()), None)
#         target = next((node['nodeId'] for node in created_node_ids if target_key in node['nodeId'].lower()), None)
#
#         if source and target:
#             # Construct relationship details based on found source and target nodes
#             relationship_details = {
#                 'source': source,
#                 'target': target,
#                 'type': node_data.get('type')
#             }
#
#             # Add an edge between the source and target nodes using the client
#             await client.add_edge(source, target, relationship_name=relationship_details['type'])



async def add_edge(client, parent_id: Optional[str], node_id: str, node_data: dict, created_node_ids):

    if node_id == "Relationship_default" and parent_id:
        # Initialize source and target variables outside the loop
        source, target = None, None

        # Assuming 'source' and 'target' are keys in node_data
        source_key = node_data.get('source', '').lower()
        target_key = node_data.get('target', '').lower()


        # Search for source and target nodes in the created_node_ids list
        for node in created_node_ids:
            node_id_lower = node['nodeId'].lower()
            if source_key in node_id_lower:
                source = node['nodeId']
                print("SOURCE FOUND", source)
            elif target_key in node_id_lower:
                target = node['nodeId']
                print("TARGET FOUND", target)

        # Check if both source and target are found
        if source and target:
            relationship_details = {
                'source': source,
                'target': target,
                'type': node_data.get('type')  # Safely access 'type' from node_data
            }
            # If you need to do something with relationship_details (e.g., add an edge), do it here
            # For demonstration, we'll just print the results
            print("RELATIONSHIP DETAILS", relationship_details)

            await client.add_edge(relationship_details['source'], relationship_details['target'], relationship_name = relationship_details['type'])




async def process_attribute(graph_client, parent_id: Optional[str], attribute: str, value: Any, created_node_ids=None):

    if created_node_ids is None:
        created_node_ids = []
    if isinstance(value, BaseModel):
        node_id = await generate_node_id(value)
        node_data = value.model_dump()

        # Use the specified default relationship for the edge between the parent node and the current node

        created_node_id = await add_node(graph_client, parent_id, node_id, node_data)

        created_node_ids.append(created_node_id)

        # await add_edge(graph_client, parent_id, node_id, node_data, relationship_data,created_node_ids)

        # Recursively process nested attributes to ensure all nodes and relationships are added to the graph
        for sub_attr, sub_val in value.__dict__.items():  # Access attributes and their values directly

            out = await process_attribute(graph_client, node_id, sub_attr, sub_val)

            created_node_ids.extend(out)

    elif isinstance(value, list) and all(isinstance(item, BaseModel) for item in value):
        # For lists of BaseModel instances, process each item in the list
        for item in value:
            out = await process_attribute(graph_client, parent_id, attribute, item, created_node_ids)
            created_node_ids.extend(out)

    return created_node_ids


async def process_attribute_edge(graph_client, parent_id: Optional[str], attribute: str, value: Any, created_node_ids=None):

    if isinstance(value, BaseModel):
        node_id = await generate_node_id(value)


        node_data = value.model_dump()
        relationship_data = {}

        await add_edge(graph_client, parent_id, node_id, node_data,created_node_ids)

        # Recursively process nested attributes to ensure all nodes and relationships are added to the graph
        for sub_attr, sub_val in value.__dict__.items():  # Access attributes and their values directly

            await process_attribute_edge(graph_client, node_id, sub_attr, sub_val, created_node_ids)

    elif isinstance(value, list) and all(isinstance(item, BaseModel) for item in value):
        # For lists of BaseModel instances, process each item in the list
        for item in value:
            await process_attribute_edge(graph_client, parent_id, attribute, item)

    return created_node_ids

async def create_dynamic(graph_model, graph_client) :
    root_id = await generate_node_id(graph_model)

    # node_data = graph_model.model_dump(exclude = {"default_relationship", "id"})
    node_data = graph_model.model_dump()

    root_id = root_id.replace(":", "_")

    _ = node_data.pop("node_id", None)

    created_node_ids = []

    out = await graph_client.add_node(root_id, node_properties = node_data)

    created_node_ids.append(out)

    for attribute_name, attribute_value in graph_model:
        ids = await process_attribute(graph_client, root_id, attribute_name, attribute_value)
        created_node_ids.extend(ids)

    flattened_and_deduplicated = list({
        item["nodeId"]: item
            # Use the 'nodeId' as the unique key in the dictionary comprehension
            for sublist in created_node_ids if sublist  # Ensure sublist is not None
            for item in sublist  # Iterate over items in the sublist
    }.values())

    for attribute_name, attribute_value in graph_model:
        ids = await process_attribute_edge(graph_client, root_id, attribute_name, attribute_value, flattened_and_deduplicated)

    return graph_client


async def create_semantic_graph(graph_model_instance, graph_client):
    # Dynamic graph creation based on the provided graph model instance
    return await create_dynamic(graph_model_instance, graph_client)



from cognee.modules.discovery import discover_directory_datasets
from cognee.infrastructure import infrastructure_config

class datasets():
    @staticmethod
    def list_datasets():
        db = infrastructure_config.get_config("database_engine")
        return db.get_datasets()

    @staticmethod
    def discover_datasets(directory_path: str):
        return list(discover_directory_datasets(directory_path).keys())

    @staticmethod
    def query_data(dataset_name: str):
        db = infrastructure_config.get_config("database_engine")
        return db.get_files_metadata(dataset_name)



import uuid
from typing import List
from qdrant_client.models import PointStruct
from cognee.infrastructure.databases.vector.get_vector_database import get_vector_database
from cognee.infrastructure.llm.openai.openai_tools import async_get_embedding_with_backoff

async def create_information_points(memory_name: str, payload: List[str]):
    vector_db = get_vector_database()

    data_points = list()
    for point in map(create_data_point, payload):
        data_points.append(await point)

    return await vector_db.create_data_points(memory_name, data_points)

async def create_data_point(data: str) -> PointStruct:
    return PointStruct(
        id = str(uuid.uuid4()),
        vector = await async_get_embedding_with_backoff(data),
        payload = {
            "raw": data,
        }
    )



""" This module contains the search function that is used to search for nodes in the graph."""
import asyncio
from enum import Enum
from typing import Dict, Any, Callable, List
from pydantic import BaseModel, field_validator
from cognee.modules.search.graph.search_adjacent import search_adjacent
from cognee.modules.search.vector.search_similarity import search_similarity
from cognee.modules.search.graph.search_categories import search_categories
from cognee.modules.search.graph.search_neighbour import search_neighbour
from cognee.modules.search.graph.search_summary import search_summary
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
from cognee.infrastructure import infrastructure_config
from cognee.utils import send_telemetry

class SearchType(Enum):
    ADJACENT = 'ADJACENT'
    SIMILARITY = 'SIMILARITY'
    CATEGORIES = 'CATEGORIES'
    NEIGHBOR = 'NEIGHBOR'
    SUMMARY = 'SUMMARY'

    @staticmethod
    def from_str(name: str):
        try:
            return SearchType[name.upper()]
        except KeyError:
            raise ValueError(f"{name} is not a valid SearchType")

class SearchParameters(BaseModel):
    search_type: SearchType
    params: Dict[str, Any]

    @field_validator('search_type', mode='before')
    def convert_string_to_enum(cls, value):
        if isinstance(value, str):
            return SearchType.from_str(value)
        return value


async def search(search_type: str, params: Dict[str, Any]) -> List:
    search_params = SearchParameters(search_type = search_type, params = params)
    return await specific_search([search_params])


async def specific_search(query_params: List[SearchParameters]) -> List:
    graph_client = await get_graph_client(infrastructure_config.get_config()["graph_engine"])
    graph = graph_client.graph

    search_functions: Dict[SearchType, Callable] = {
        SearchType.ADJACENT: search_adjacent,
        SearchType.SIMILARITY: search_similarity,
        SearchType.CATEGORIES: search_categories,
        SearchType.NEIGHBOR: search_neighbour,
        SearchType.SUMMARY: search_summary
    }

    results = []
    search_tasks = []

    for search_param in query_params:
        search_func = search_functions.get(search_param.search_type)
        if search_func:
            # Schedule the coroutine for execution and store the task
            task = search_func(**search_param.params, graph = graph)
            search_tasks.append(task)

    # Use asyncio.gather to run all scheduled tasks concurrently
    search_results = await asyncio.gather(*search_tasks)

    # Update the results set with the results from all tasks
    results.extend(search_results)

    send_telemetry( "COGNEE_SEARCH")

    return results



if __name__ == "__main__":
    async def main():
        # Assuming 'graph' is your graph object, obtained from somewhere
        search_type = 'CATEGORIES'
        params = {'query': 'Ministarstvo', 'other_param': {"node_id": "LLM_LAYER_SUMMARY:DOCUMENT:881ecb36-2819-54c3-8147-ed80293084d6"}}

        results = await search(search_type, params)
        print(results)

    # Run the async main function
    asyncio.run(main())
# if __name__ == "__main__":
#     import asyncio

#     query_params = {
#         SearchType.SIMILARITY: {'query': 'your search query here'}
#     }
#     async def main():
#         graph_client = get_graph_client(GraphDBType.NETWORKX)

#         await graph_client.load_graph_from_file()
#         graph = graph_client.graph
#         results = await search(graph, query_params)
#         print(results)

#     asyncio.run(main())



from typing import List
from datetime import datetime, timezone
from sqlalchemy.orm import relationship, MappedColumn, Mapped
from sqlalchemy import Column, String, DateTime, UUID, Text, JSON
from cognee.infrastructure.databases.relational import ModelBase
from .DatasetData import DatasetData

class Data(ModelBase):
    __tablename__ = "data"

    id = Column(UUID, primary_key = True)
    name = Column(String, nullable = True)
    description = Column(Text, nullable = True)
    raw_data_location = Column(String)
    meta_data: Mapped[dict] = MappedColumn(type_ = JSON) # metadata is reserved word

    created_at = Column(DateTime, default = datetime.now(timezone.utc))
    updated_at = Column(DateTime, onupdate = datetime.now(timezone.utc))

    datasets: Mapped[List["Dataset"]] = relationship(
        secondary = DatasetData.__tablename__,
        back_populates = "data"
    )



import duckdb

class DuckDBAdapter():
    def __init__(self, db_path: str, db_name: str):
        db_location = db_path + "/" + db_name

        self.get_connection = lambda: duckdb.connect(db_location)

    def get_datasets(self):
        with self.get_connection() as connection:
            tables = connection.sql("SELECT DISTINCT schema_name FROM duckdb_tables();").to_df().to_dict("list")

        return list(
            filter(
                lambda table_name: table_name.endswith("staging") is False,
                tables["schema_name"]
            )
        )

    def get_files_metadata(self, dataset_name: str):
        with self.get_connection() as connection:
            return connection.sql(f"SELECT id, name, file_path, extension, mime_type, keywords FROM {dataset_name}.file_metadata;").to_df().to_dict("records")


    def load_cognify_data(self, data):
        with self.get_connection() as connection:
            # Ensure the "cognify" table exists
            connection.execute("""
                CREATE TABLE IF NOT EXISTS cognify (
                    document_id STRING,
                    layer_id STRING,
                  created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    updated_at TIMESTAMP DEFAULT NULL,
                    processed BOOLEAN DEFAULT FALSE,
                    document_id_target STRING NULL
                );
            """)

        # Prepare the insert statement
        insert_query = """
            INSERT INTO cognify (document_id, layer_id)
            VALUES (?, ?);
        """

        # Insert each record into the "cognify" table
        for record in data:
            with self.get_connection() as connection:
                connection.execute(insert_query, [
                    record.get("document_id"),
                    record.get("layer_id")
                ])

    def fetch_cognify_data(self, excluded_document_id: str):
        # SQL command to create the "cognify" table with the specified columns
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS cognify (
            document_id STRING,
            layer_id STRING,
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP DEFAULT NULL,
            processed BOOLEAN DEFAULT FALSE,
            document_id_target STRING NULL
        );
        """
        with self.get_connection() as connection:
            # Execute the SQL command to create the table
            connection.execute(create_table_sql)

        # SQL command to select data from the "cognify" table
        select_data_sql = f"SELECT document_id, layer_id, created_at, updated_at, processed FROM cognify WHERE document_id != '{excluded_document_id}' AND processed = FALSE;"

        with self.get_connection() as connection:
            # Execute the query and fetch the results
            records = connection.sql(select_data_sql).to_df().to_dict("records")

        # If records are fetched, update the "processed" column to "True"
        if records:
            # Fetching document_ids from the records to update the "processed" column
            document_ids = tuple(record["document_id"] for record in records)
            # SQL command to update the "processed" column to "True" for fetched records
            update_data_sql = f"UPDATE cognify SET processed = TRUE WHERE document_id IN {document_ids};"

            with self.get_connection() as connection:
                # Execute the update query
                connection.execute(update_data_sql)

        # Return the fetched records
        return records


    def delete_cognify_data(self):
        # SQL command to create the "cognify" table with the specified columns
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS cognify (
            document_id STRING,
            layer_id STRING,
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP DEFAULT NULL,
            processed BOOLEAN DEFAULT FALSE,
            document_id_target STRING NULL
        );
        """

        with self.get_connection() as connection:
            # Execute the SQL command to create the table
            connection.execute(create_table_sql)

        with self.get_connection() as connection:
            # SQL command to select data from the "cognify" table
            select_data_sql = "DELETE FROM cognify;"
            connection.sql(select_data_sql)
            drop_data_sql = "DROP TABLE cognify;"
            connection.sql(drop_data_sql)



import os
import asyncio
from typing import Callable
from sqlalchemy.inspection import inspect
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker, AsyncEngine, AsyncSession, async_scoped_session
from sqlalchemy.future import select
from cognee.infrastructure.files.storage.LocalStorage import LocalStorage
from ..DatabaseEngine import DatabaseEngine
from ..ModelBase import ModelBase
from ..utils import with_rollback

class SqliteEngine(DatabaseEngine):
    db_path: str = None
    db_name: str = None
    engine: AsyncEngine = None
    session_maker: Callable[[], async_scoped_session[AsyncSession]] = None
    is_db_done: bool = False

    def __init__(self, db_path: str, db_name: str):
        self.db_path = db_path
        self.db_name = db_name
        self.db_location = db_path + "/" + db_name
        self.engine = create_async_engine(
            f"sqlite+aiosqlite:///{self.db_location}",
            pool_recycle = 3600,
            echo = False
        )
        self.session_maker = lambda: async_scoped_session(
            async_sessionmaker(
                bind = self.engine,
                class_ = AsyncSession
            ),
            scopefunc = asyncio.current_task
        )

    async def ensure_tables(self):
        if not self.database_exists(self.db_name):
            self.create_database(self.db_name)

            await self.create_tables()

            self.is_db_done = True

            return True

    def database_exists(self, db_name: str) -> bool:
        return os.path.exists(self.db_path + "/" + db_name)

    def create_database(self, db_name: str):
        LocalStorage.ensure_directory_exists(self.db_path)

        with open(self.db_path + "/" + db_name, mode = "w+", encoding = "utf-8") as file:
            file.write("")

    def drop_database(self, db_name: str):
        os.remove(self.db_location)

    async def table_exists(self, table_name: str) -> bool:
        return inspect(self.engine).has_table(table_name)

    async def create_tables(self):
        async with self.engine.begin() as connection:
            return await connection.run_sync(ModelBase.metadata.create_all)

    async def create(self, data):
        async with with_rollback(self.session_maker()) as session:
            session.add(data)

    async def query(self, query_term):
        async with with_rollback(self.session_maker()) as session:
            return await session.execute(query_term)

    async def query_entity(self, entity):
        async with with_rollback(self.session_maker()) as session:
            return await session.execute(
                select(type(entity))
                    .where(type(entity).id == entity.id)
            )

    async def update(self, data_update_fn):
        async with with_rollback(self.session_maker()):
            data_update_fn()



from cognee.infrastructure.databases.relational.get_database import get_database

def register_memory_for_user(user_id: str, memory_name: str):
    return get_database().add_memory(user_id, memory_name)



from typing import List
import dspy
import nltk
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from cognee.config import Config
from cognee.shared.data_models import KnowledgeGraph, Node, Edge
from cognee.utils import num_tokens_from_string, trim_text_to_max_tokens

config = Config()
config.load()

# """Instructions:
# You are a top-tier algorithm designed for extracting information from text in structured formats to build a knowledge graph.
# - **Nodes** represent entities and concepts. They're akin to Wikipedia nodes.
# - **Edges** represent relationships between concepts. They're akin to Wikipedia links.
# Extract as much information as you can from the text and build a detailed knowledge graph.
# If question is provided, make sure that the information to answer the question is present in the graph."""

class GraphFromText(dspy.Signature):
    """Instructions:
    You are a top-tier algorithm designed for extracting information from text in structured formats to build a knowledge graph.
    - **Nodes** represent entities and concepts, akin to Wikipedia nodes.
    - **Edges** represent relationships between entities and concepts, akin to Wikipedia hyperlinks.
    Extract information from the text and build a detailed knowledge graph."""

    text: str = dspy.InputField()
    graph: KnowledgeGraph = dspy.OutputField()


def are_all_nodes_and_edges_valid(graph: KnowledgeGraph) -> bool:
    return all([getattr(node, "entity_type", "").strip() != "" for node in graph.nodes]) and \
        all([getattr(node, "entity_name", "").strip() != "" for node in graph.nodes]) and \
        all([getattr(edge, "relationship_name", "").strip() != "" for edge in graph.edges])

def is_node_connected(node: Node, edges: List[Edge]) -> bool:
    return any([(edge.source_node_id == node.id or edge.target_node_id == node.id) for edge in edges])

def are_all_nodes_connected(graph: KnowledgeGraph) -> bool:
    return all([is_node_connected(node, graph.edges) for node in graph.nodes])


class ExtractKnowledgeGraph(dspy.Module):
    def __init__(self, lm = dspy.OpenAI(model = config.openai_model, api_key = config.openai_key, model_type = "chat", max_tokens = 4096)):
        super().__init__()
        self.lm = lm
        dspy.settings.configure(lm=self.lm)
        self.generate_graph = dspy.TypedChainOfThought(GraphFromText)
        nltk.download('stopwords')

    def forward(self, context: str, question: str):
        context = remove_stop_words(context)
        context = trim_text_to_max_tokens(context, 1500, config.openai_model)
      
        with dspy.context(lm = self.lm):
            graph = self.generate_graph(text = context).graph

            not_valid_nodes_or_edges_message = """
                All nodes must contain "entity_name".
                All edges must contain "relationship_name".
                Please add mandatory fields to nodes and edges."""

            dspy.Suggest(are_all_nodes_and_edges_valid(graph), not_valid_nodes_or_edges_message)

            # not_connected_graph_message = """
            #     Output must be a graph that has all nodes connected to it.
            #     Please find a relation and connect nodes or remove them."""

            # dspy.Suggest(are_all_nodes_connected(graph), not_connected_graph_message)

        return dspy.Prediction(context = context, graph = graph)


def remove_stop_words(text):
    stop_words = set(stopwords.words("english"))
    word_tokens = word_tokenize(text)
    filtered_text = [word for word in word_tokens if word.lower() not in stop_words]
    return " ".join(filtered_text)

#
# if __name__ == "__main__":
#     gpt_4_turbo = dspy.OpenAI(model="gpt-4", max_tokens=4000, api_key=config.openai_key, model_type="chat")
#     dspy.settings.configure(lm=gpt_4_turbo)


#     extract_knowledge_graph = ExtractKnowledgeGraph(lm=gpt_4_turbo)
#     # graph_text = extract_knowledge_graph("cognitive_layer", "text")
#     graph = extract_knowledge_graph("analysis_layer", """A large language model (LLM) is a language model notable for its ability to achieve general-purpose language generation and other natural language processing tasks such as classification. LLMs acquire these abilities by learning statistical relationships from text documents during a computationally intensive self-supervised and semi-supervised training process. LLMs can be used for text generation, a form of generative AI, by taking an input text and repeatedly predicting the next token or word.
# LLMs are artificial neural networks. The largest and most capable, as of March 2024""", question="What is a large language model?")
#     print("GPT4 History:", gpt_4_turbo.inspect_history(n=1))
#     print(graph)
#
